from collections.abc import Sequence

import numpy

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import ObjectGeometry, ObjectGeometryProvider
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.probe import ProbeGeometry, ProbeGeometryProvider
from ptychodus.api.product import (
    ELECTRON_VOLT_J,
    LIGHT_SPEED_M_PER_S,
    PLANCK_CONSTANT_J_PER_HZ,
)
from ptychodus.api.probe_positions import ProbePosition

from ..diffraction import PatternSizer
from .metadata import MetadataRepositoryItem
from .probe_positions import ProbePositionsRepositoryItem


class ProductGeometry(ProbeGeometryProvider, ObjectGeometryProvider, Observable, Observer):
    def __init__(
        self,
        pattern_sizer: PatternSizer,
        metadata_item: MetadataRepositoryItem,
        scan_item: ProbePositionsRepositoryItem,
    ) -> None:
        super().__init__()
        self._pattern_sizer = pattern_sizer
        self._metadata_item = metadata_item
        self._scan_item = scan_item

        self._pattern_sizer.add_observer(self)
        self._metadata_item.add_observer(self)
        self._scan_item.add_observer(self)

    @property
    def probe_photon_count(self) -> float:
        return self._metadata_item.probe_photon_count.get_value()

    @property
    def probe_energy_J(self) -> float:  # noqa: N802
        return self._metadata_item.probe_energy_eV.get_value() * ELECTRON_VOLT_J

    @property
    def probe_wavelength_m(self) -> float:
        hc_Jm = PLANCK_CONSTANT_J_PER_HZ * LIGHT_SPEED_M_PER_S  # noqa: N806

        try:
            return hc_Jm / self.probe_energy_J
        except ZeroDivisionError:
            return 0.0

    @property
    def probe_wavelengths_per_m(self) -> float:
        """wavenumber"""
        return 1.0 / self.probe_wavelength_m

    @property
    def probe_radians_per_m(self) -> float:
        """angular wavenumber"""
        return 2.0 * numpy.pi / self.probe_wavelength_m

    @property
    def probe_photons_per_s(self) -> float:
        try:
            return self.probe_photon_count / self._metadata_item.exposure_time_s.get_value()
        except ZeroDivisionError:
            return 0.0

    @property
    def probe_power_W(self) -> float:  # noqa: N802
        return self.probe_energy_J * self.probe_photons_per_s

    @property
    def num_scan_points(self) -> int:
        return len(self._scan_item.get_probe_positions())

    @property
    def detector_distance_m(self) -> float:
        return self._metadata_item.detector_distance_m.get_value()

    @property
    def _lambda_z_m2(self) -> float:
        return self.probe_wavelength_m * self.detector_distance_m

    @property
    def object_plane_pixel_width_m(self) -> float:
        return self._lambda_z_m2 / self._pattern_sizer.get_processed_width_m()

    @property
    def object_plane_pixel_height_m(self) -> float:
        return self._lambda_z_m2 / self._pattern_sizer.get_processed_height_m()

    def get_detector_pixel_geometry(self):
        return self._pattern_sizer.get_processed_pixel_geometry()

    def get_object_plane_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self.object_plane_pixel_width_m,
            height_m=self.object_plane_pixel_height_m,
        )

    @property
    def fresnel_number(self) -> float:
        width_m = self._pattern_sizer.get_processed_width_m()
        height_m = self._pattern_sizer.get_processed_height_m()
        area_m2 = width_m * height_m
        return area_m2 / self._lambda_z_m2

    @property
    def _detector_numerical_aperture_sq(self) -> float:
        two_z_m = 2 * self.detector_distance_m
        NA_x = self._pattern_sizer.get_processed_width_m() / two_z_m  # noqa: N806
        NA_y = self._pattern_sizer.get_processed_height_m() / two_z_m  # noqa: N806
        return NA_x * NA_y

    @property
    def detector_numerical_aperture(self) -> float:
        return numpy.sqrt(self._detector_numerical_aperture_sq)

    @property
    def depth_of_field_m(self) -> float:
        return self.probe_wavelength_m / self._detector_numerical_aperture_sq

    def get_probe_geometry(self) -> ProbeGeometry:
        extent = self._pattern_sizer.get_processed_image_extent()
        return ProbeGeometry(
            width_px=extent.width_px,
            height_px=extent.height_px,
            pixel_width_m=self.object_plane_pixel_width_m,
            pixel_height_m=self.object_plane_pixel_height_m,
        )

    def is_probe_geometry_valid(self, geometry: ProbeGeometry) -> bool:
        expected = self.get_probe_geometry()
        width_is_valid = geometry.pixel_width_m > 0.0 and geometry.width_m == expected.width_m
        height_is_valid = geometry.pixel_height_m > 0.0 and geometry.height_m == expected.height_m
        return width_is_valid and height_is_valid

    def get_probe_positions(self) -> Sequence[ProbePosition]:
        return self._scan_item.get_probe_positions()

    def get_object_geometry(self) -> ObjectGeometry:
        probe_geometry = self.get_probe_geometry()
        width_m = probe_geometry.width_m
        height_m = probe_geometry.height_m
        center_x_m = 0.0
        center_y_m = 0.0

        scan_bbox = self._scan_item.get_bounding_box()

        if scan_bbox is not None:
            width_m += scan_bbox.width_m
            height_m += scan_bbox.height_m
            center_x_m = scan_bbox.center_x_m
            center_y_m = scan_bbox.center_y_m

        pixel_width_m = self.object_plane_pixel_width_m
        width_px = width_m / pixel_width_m if pixel_width_m > 0.0 else 0.0

        pixel_height_m = self.object_plane_pixel_height_m
        height_px = height_m / pixel_height_m if pixel_height_m > 0.0 else 0.0

        return ObjectGeometry(
            width_px=int(numpy.ceil(width_px)),
            height_px=int(numpy.ceil(height_px)),
            pixel_width_m=self.object_plane_pixel_width_m,
            pixel_height_m=self.object_plane_pixel_height_m,
            center_x_m=center_x_m,
            center_y_m=center_y_m,
        )

    def is_object_geometry_valid(self, geometry: ObjectGeometry) -> bool:
        expected_geometry = self.get_object_geometry()
        pixel_size_is_valid = geometry.pixel_width_m > 0.0 and geometry.pixel_height_m > 0.0
        return pixel_size_is_valid and geometry.contains(expected_geometry)

    def _update(self, observable: Observable) -> None:
        if observable is self._metadata_item:
            self.notify_observers()
        elif observable is self._scan_item:
            self.notify_observers()
        elif observable is self._pattern_sizer:
            self.notify_observers()



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from decimal import Decimal
from pathlib import Path
from typing import Final
from uuid import UUID
import logging

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIntValidator, QValidator
from PyQt5.QtWidgets import (
    QAbstractButton,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from ptychodus.api.geometry import Interval
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import (
    BooleanParameter,
    IntegerParameter,
    Parameter,
    PathParameter,
    RealParameter,
    StringParameter,
    UUIDParameter,
)

from ..view.widgets import AngleWidget, DecimalLineEdit, DecimalSlider, LengthWidget
from .data import FileDialogFactory

logger = logging.getLogger(__name__)


class ParameterViewController(ABC):
    @abstractmethod
    def get_widget(self) -> QWidget:
        pass


class CheckableGroupBoxParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: BooleanParameter, title: str, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QGroupBox(title)
        self._widget.setCheckable(True)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.toggled.connect(parameter.set_value)
        self._parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_model_to_view(self) -> None:
        self._widget.setChecked(self._parameter.get_value())

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class CheckBoxParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: BooleanParameter, text: str, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QCheckBox(text)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.toggled.connect(parameter.set_value)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_model_to_view(self) -> None:
        self._widget.setChecked(self._parameter.get_value())

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class SpinBoxParameterViewController(ParameterViewController, Observer):
    MAX_INT: Final[int] = 0x7FFFFFFF

    def __init__(self, parameter: IntegerParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QSpinBox()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.valueChanged.connect(parameter.set_value)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_model_to_view(self) -> None:
        minimum = self._parameter.get_minimum()
        maximum = self._parameter.get_maximum()

        if minimum is None:
            raise ValueError('Minimum not provided!')
        else:
            self._widget.blockSignals(True)

            if maximum is None:
                self._widget.setRange(minimum, SpinBoxParameterViewController.MAX_INT)
            else:
                self._widget.setRange(minimum, maximum)

            self._widget.setValue(self._parameter.get_value())
            self._widget.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class ComboBoxParameterViewController(ParameterViewController, Observer):
    def __init__(
        self, parameter: Parameter[str], items: Iterable[str], *, tool_tip: str = ''
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QComboBox()

        for item in items:
            self._widget.addItem(item)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.textActivated.connect(parameter.set_value)
        parameter.add_observer(self)

    def get_widget(self) -> QComboBox:
        return self._widget

    def __sync_model_to_view(self) -> None:
        self._widget.setCurrentText(self._parameter.get_value())

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class LineEditParameterViewController(ParameterViewController, Observer):
    def __init__(
        self, parameter: StringParameter, validator: QValidator | None = None, *, tool_tip: str = ''
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QLineEdit()

        if validator is not None:
            self._widget.setValidator(validator)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.editingFinished.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self) -> None:
        self._parameter.set_value(self._widget.text())

    def __sync_model_to_view(self) -> None:
        self._widget.setText(self._parameter.get_value())

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class PathParameterViewController(ParameterViewController, Observer):
    def __init__(
        self,
        parameter: PathParameter,
        file_dialog_factory: FileDialogFactory,
        *,
        caption: str,
        name_filters: Sequence[str] | None,
        mime_type_filters: Sequence[str] | None,
        selected_name_filter: str | None,
        tool_tip: str,
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._file_dialog_factory = file_dialog_factory
        self._caption = caption
        self._name_filters = name_filters
        self._mime_type_filters = mime_type_filters
        self._selected_name_filter = selected_name_filter
        self._line_edit = QLineEdit()
        self._browse_button = QPushButton('Browse')
        self._widget = QWidget()

        if tool_tip:
            self._line_edit.setToolTip(tool_tip)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._line_edit)
        layout.addWidget(self._browse_button)
        self._widget.setLayout(layout)

        self.__sync_model_to_view()
        parameter.add_observer(self)
        self._line_edit.editingFinished.connect(self.__sync_path_to_model)

    @classmethod
    def create_file_opener(
        cls,
        parameter: PathParameter,
        file_dialog_factory: FileDialogFactory,
        *,
        caption: str = 'Open File',
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
        tool_tip: str = '',
    ) -> PathParameterViewController:
        view_controller = cls(
            parameter,
            file_dialog_factory,
            caption=caption,
            name_filters=name_filters,
            mime_type_filters=mime_type_filters,
            selected_name_filter=selected_name_filter,
            tool_tip=tool_tip,
        )
        view_controller._browse_button.clicked.connect(view_controller._choose_file_to_open)
        return view_controller

    @classmethod
    def create_file_saver(
        cls,
        parameter: PathParameter,
        file_dialog_factory: FileDialogFactory,
        *,
        caption: str = 'Save File',
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
        tool_tip: str = '',
    ) -> PathParameterViewController:
        view_controller = cls(
            parameter,
            file_dialog_factory,
            caption=caption,
            name_filters=name_filters,
            mime_type_filters=mime_type_filters,
            selected_name_filter=selected_name_filter,
            tool_tip=tool_tip,
        )
        view_controller._browse_button.clicked.connect(view_controller._choose_file_to_save)
        return view_controller

    @classmethod
    def create_directory_chooser(
        cls,
        parameter: PathParameter,
        file_dialog_factory: FileDialogFactory,
        *,
        caption: str = 'Choose Directory',
        tool_tip: str = '',
    ) -> PathParameterViewController:
        view_controller = cls(
            parameter,
            file_dialog_factory,
            caption=caption,
            name_filters=None,
            mime_type_filters=None,
            selected_name_filter=None,
            tool_tip=tool_tip,
        )
        view_controller._browse_button.clicked.connect(view_controller._choose_directory)
        return view_controller

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_path_to_model(self) -> None:
        path = Path(self._line_edit.text())
        self._parameter.set_value(path)

    def _choose_file_to_open(self) -> None:
        path, _ = self._file_dialog_factory.get_open_file_path(
            self._widget,
            self._caption,
            self._name_filters,
            self._mime_type_filters,
            self._selected_name_filter,
        )

        if path:
            self._parameter.set_value(path)

    def _choose_file_to_save(self) -> None:
        path, _ = self._file_dialog_factory.get_save_file_path(
            self._widget,
            self._caption,
            self._name_filters,
            self._mime_type_filters,
            self._selected_name_filter,
        )

        if path:
            self._parameter.set_value(path)

    def _choose_directory(self) -> None:
        path = self._file_dialog_factory.get_existing_directory_path(self._widget, self._caption)

        if path:
            self._parameter.set_value(path)

    def __sync_model_to_view(self) -> None:
        path = self._parameter.get_value()
        self._line_edit.setText(str(path))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class IntegerLineEditParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: IntegerParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QLineEdit()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        validator = QIntValidator()
        bottom = parameter.get_minimum()
        top = parameter.get_maximum()

        if bottom is not None:
            validator.setBottom(bottom)

        if top is not None:
            validator.setTop(top)

        self._widget.setValidator(validator)

        self.__sync_model_to_view()
        self._widget.editingFinished.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self) -> None:
        text = self._widget.text()

        try:
            value = int(text)
        except ValueError:
            logger.warning(f'Failed to convert "{text}" to int!')
        else:
            self._parameter.set_value(value)

    def __sync_model_to_view(self) -> None:
        self._widget.setText(str(self._parameter.get_value()))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class DecimalLineEditParameterViewController(ParameterViewController, Observer):
    def __init__(
        self, parameter: RealParameter, *, is_signed: bool = False, tool_tip: str = ''
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = DecimalLineEdit.create_instance(is_signed=is_signed)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.value_changed.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self, value: Decimal) -> None:
        self._parameter.set_value(float(value))

    def __sync_model_to_view(self) -> None:
        self._widget.set_value(Decimal(str(self._parameter.get_value())))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class DecimalSliderParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: RealParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = DecimalSlider.create_instance(Qt.Orientation.Horizontal)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.value_changed.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self, value: Decimal) -> None:
        self._parameter.set_value(float(value))

    def __sync_model_to_view(self) -> None:
        minimum = self._parameter.get_minimum()
        maximum = self._parameter.get_maximum()

        if minimum is None or maximum is None:
            raise ValueError('Range not provided!')
        else:
            value = Decimal(str(self._parameter.get_value()))
            range_ = Interval[Decimal](Decimal(str(minimum)), Decimal(str(maximum)))
            self._widget.set_value_and_range(value, range_)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class UUIDLineEditParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: UUIDParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QLineEdit()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.editingFinished.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self) -> None:
        self._parameter.set_value(UUID(self._widget.text()))

    def __sync_model_to_view(self) -> None:
        self._widget.setText(str(self._parameter.get_value()))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class LengthWidgetParameterViewController(ParameterViewController, Observer):
    def __init__(
        self, parameter: RealParameter, *, is_signed: bool = False, tool_tip: str = ''
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = LengthWidget(is_signed=is_signed)

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.length_changed.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self, value: Decimal) -> None:
        self._parameter.set_value(float(value))

    def __sync_model_to_view(self) -> None:
        self._widget.set_length_m(Decimal(str(self._parameter.get_value())))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class AngleWidgetParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: RealParameter, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = AngleWidget()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.angle_changed.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self, value: Decimal) -> None:
        self._parameter.set_value(float(value))

    def __sync_model_to_view(self) -> None:
        self._widget.set_angle_in_turns(Decimal(str(self._parameter.get_value())))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class ParameterWidget(QWidget):
    def __init__(
        self, view_controllers: Sequence[ParameterViewController], parent: QWidget | None = None
    ) -> None:
        super().__init__(parent)
        self._view_controllers = view_controllers


class ParameterDialog(QDialog):
    def __init__(
        self,
        view_controllers: Sequence[ParameterViewController],
        button_box: QDialogButtonBox,
        parent: QWidget | None,
    ) -> None:
        super().__init__(parent)
        self._view_controllers = view_controllers
        self._button_box = button_box

        button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        button_box.clicked.connect(self._handle_button_box_clicked)

    def _handle_button_box_clicked(self, button: QAbstractButton) -> None:
        # TODO remove observers from viewControllers

        if self._button_box.buttonRole(button) == QDialogButtonBox.ButtonRole.AcceptRole:
            self.accept()
        else:
            self.reject()


class ParameterViewBuilder:
    def __init__(self, file_dialog_factory: FileDialogFactory | None = None) -> None:
        self._file_dialog_factory = file_dialog_factory
        self._view_controllers_top: list[ParameterViewController] = list()
        self._view_controllers: dict[tuple[str, str], ParameterViewController] = dict()
        self._view_controllers_bottom: list[ParameterViewController] = list()

    def add_check_box(
        self,
        parameter: BooleanParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = CheckBoxParameterViewController(parameter, '', tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_combo_box(
        self,
        parameter: StringParameter,
        items: Iterable[str],
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = ComboBoxParameterViewController(parameter, items, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_file_opener(
        self,
        parameter: PathParameter,
        label: str,
        *,
        caption: str = 'Open File',
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        if self._file_dialog_factory is None:
            raise ValueError('Cannot add file chooser without FileDialogFactory!')
        else:
            view_controller = PathParameterViewController.create_file_opener(
                parameter,
                self._file_dialog_factory,
                caption=caption,
                name_filters=name_filters,
                mime_type_filters=mime_type_filters,
                selected_name_filter=selected_name_filter,
                tool_tip=tool_tip,
            )
            self.add_view_controller(view_controller, label, group=group)

    def add_file_saver(
        self,
        parameter: PathParameter,
        label: str,
        *,
        caption: str = 'Save File',
        name_filters: Sequence[str] | None = None,
        mime_type_filters: Sequence[str] | None = None,
        selected_name_filter: str | None = None,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        if self._file_dialog_factory is None:
            raise ValueError('Cannot add file chooser without FileDialogFactory!')
        else:
            view_controller = PathParameterViewController.create_file_saver(
                parameter,
                self._file_dialog_factory,
                caption=caption,
                name_filters=name_filters,
                mime_type_filters=mime_type_filters,
                selected_name_filter=selected_name_filter,
                tool_tip=tool_tip,
            )
            self.add_view_controller(view_controller, label, group=group)

    def add_directory_chooser(
        self, parameter: PathParameter, label: str, *, tool_tip: str = '', group: str = ''
    ) -> None:
        if self._file_dialog_factory is None:
            raise ValueError('Cannot add directory chooser without FileDialogFactory!')
        else:
            view_controller = PathParameterViewController.create_directory_chooser(
                parameter,
                self._file_dialog_factory,
                tool_tip=tool_tip,
            )
            self.add_view_controller(view_controller, label, group=group)

    def add_spin_box(
        self,
        parameter: IntegerParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = SpinBoxParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_line_edit(
        self,
        parameter: StringParameter,
        label: str,
        *,
        validator: QValidator | None = None,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = LineEditParameterViewController(
            parameter, validator=validator, tool_tip=tool_tip
        )
        self.add_view_controller(view_controller, label, group=group)

    def add_integer_line_edit(
        self,
        parameter: IntegerParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = IntegerLineEditParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_decimal_line_edit(
        self,
        parameter: RealParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = DecimalLineEditParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_decimal_slider(
        self,
        parameter: RealParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = DecimalSliderParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_uuid_line_edit(
        self,
        parameter: UUIDParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = UUIDLineEditParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_length_widget(
        self,
        parameter: RealParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = LengthWidgetParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_angle_widget(
        self,
        parameter: RealParameter,
        label: str,
        *,
        tool_tip: str = '',
        group: str = '',
    ) -> None:
        view_controller = AngleWidgetParameterViewController(parameter, tool_tip=tool_tip)
        self.add_view_controller(view_controller, label, group=group)

    def add_view_controller_to_top(self, view_controller: ParameterViewController) -> None:
        self._view_controllers_top.append(view_controller)

    def add_view_controller(
        self,
        view_controller: ParameterViewController,
        label: str,
        *,
        group: str = '',
    ) -> None:
        self._view_controllers[group, label] = view_controller

    def add_view_controller_to_bottom(self, view_controller: ParameterViewController) -> None:
        self._view_controllers_bottom.append(view_controller)

    def _build_layout(self, *, add_stretch: bool) -> QVBoxLayout:
        group_dict: dict[str, QFormLayout] = dict()

        for (group_name, widget_label), vc in self._view_controllers.items():
            try:
                form_layout = group_dict[group_name]
            except KeyError:
                form_layout = QFormLayout()
                group_dict[group_name] = form_layout

            form_layout.addRow(widget_label, vc.get_widget())

        layout = QVBoxLayout()

        for view_controller in self._view_controllers_top:
            layout.addWidget(view_controller.get_widget())

        for group_name, group_layout in group_dict.items():
            if group_name:
                group_box = QGroupBox(group_name)
                group_box.setLayout(group_layout)
                layout.addWidget(group_box)
            elif group_layout.count() > 0:
                layout.addLayout(group_layout)

        for view_controller in self._view_controllers_bottom:
            layout.addWidget(view_controller.get_widget())

        if add_stretch:
            layout.addStretch()

        return layout

    def _flush_view_controllers(self) -> Sequence[ParameterViewController]:
        view_controllers: list[ParameterViewController] = list()
        view_controllers.extend(self._view_controllers_top)
        view_controllers.extend(self._view_controllers.values())
        view_controllers.extend(self._view_controllers_bottom)

        self._view_controllers_top.clear()
        self._view_controllers.clear()
        self._view_controllers_bottom.clear()

        return view_controllers

    def build_widget(self) -> QWidget:
        layout = self._build_layout(add_stretch=True)

        widget = ParameterWidget(self._flush_view_controllers())
        widget.setLayout(layout)

        return widget

    def build_dialog(self, window_title: str, parent: QWidget | None) -> QDialog:
        button_box = QDialogButtonBox()
        layout = self._build_layout(add_stretch=True)
        layout.addWidget(button_box)

        dialog = ParameterDialog(self._flush_view_controllers(), button_box, parent)
        dialog.setLayout(layout)
        dialog.setWindowTitle(window_title)

        return dialog



from pathlib import Path
from typing import Final
import csv
import logging

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
    ProbePositionParseError,
)

logger = logging.getLogger(__name__)


class LYNXSoftGlueZynqPositionFileReader(ProbePositionFileReader):
    SIMPLE_NAME: Final[str] = 'APS_LYNX_SoftGlueZynq'
    DISPLAY_NAME: Final[str] = 'APS 31-ID-E LYNX SoftGlueZynq Files (*.dat)'
    ONE_MICRON_M: Final[float] = 1.0e-6

    EXPECTED_HEADER_RAW: Final[list[str]] = [
        'DataPoint',
        'x_st_fzp',
        'y_st_fzp',
        'ckUser_Clk_Count',
        'Detector_Count',
    ]

    EXPECTED_HEADER_PROCESSED: Final[list[str]] = [
        'Detector_Count',
        'Average_x_st_fzp',
        'Stdev_x_st_fzp',
        'Average_y_st_fzp',
        'Stdev_y_st_fzp',
    ]

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()
        scan_name = self.SIMPLE_NAME

        with file_path.open(newline='') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=' ')
            csv_iterator = iter(csv_reader)

            title_row = next(csv_iterator)

            try:
                scan_name = ' '.join(title_row).split(',', maxsplit=1)[0]
            except IndexError:
                raise ProbePositionParseError('Bad scan name!')

            column_header_row = next(csv_iterator)

            if column_header_row == LYNXSoftGlueZynqPositionFileReader.EXPECTED_HEADER_RAW:
                logger.debug(f'Reading raw scan positions for "{scan_name}"...')
                X = 1  # noqa: N806
                Y = 2  # noqa: N806
                DETECTOR_COUNT = 4  # noqa: N806
            elif column_header_row == LYNXSoftGlueZynqPositionFileReader.EXPECTED_HEADER_PROCESSED:
                logger.debug(f'Reading processed scan positions for "{scan_name}"...')
                DETECTOR_COUNT = 0  # noqa: N806
                X = 1  # noqa: N806
                Y = 3  # noqa: N806
            else:
                raise ProbePositionParseError(
                    f'Bad LYNX SoftGlueZynq header!\nFound:    {column_header_row}\n'
                )

            for row in csv_iterator:
                if row[0].startswith('#'):
                    continue

                if len(row) != len(column_header_row):
                    raise ProbePositionParseError('Bad number of columns!')

                point = ProbePosition(
                    int(row[DETECTOR_COUNT]),
                    -float(row[X]) * self.ONE_MICRON_M,
                    -float(row[Y]) * self.ONE_MICRON_M,
                )
                point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        LYNXSoftGlueZynqPositionFileReader(),
        simple_name=LYNXSoftGlueZynqPositionFileReader.SIMPLE_NAME,
        display_name=LYNXSoftGlueZynqPositionFileReader.DISPLAY_NAME,
    )



"""Observer pattern for propagating changes across ptychodus components."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

__all__ = [
    'Observer',
    'Observable',
    'SequenceObserver',
    'ObservableSequence',
]

T = TypeVar('T')


class Observer(ABC):
    """Abstract observer that receives change notifications from an Observable."""

    @abstractmethod
    def _update(self, observable: Observable) -> None:
        pass


class Observable:
    """Subject that maintains a list of observers and dispatches change notifications."""

    def __init__(self) -> None:
        self._observer_list: list[Observer] = list()
        self._block_notifications = False
        self._pending_notify = False

    def add_observer(self, observer: Observer) -> None:
        if observer not in self._observer_list:
            self._observer_list.append(observer)

    def remove_observer(self, observer: Observer) -> None:
        try:
            self._observer_list.remove(observer)
        except ValueError:
            pass

    def block_notifications(self, block: bool) -> None:
        self._block_notifications = block

        if self._pending_notify:
            self.notify_observers()
            self._pending_notify = False

    def notify_observers(self) -> None:
        if self._block_notifications:
            self._pending_notify = True
            return

        for observer in self._observer_list:
            observer._update(self)


class SequenceObserver(Generic[T], ABC):
    """Observer for fine-grained notifications about insertions, changes, and removals in a sequence."""

    @abstractmethod
    def handle_item_inserted(self, index: int, item: T) -> None:
        pass

    @abstractmethod
    def handle_item_changed(self, index: int, item: T) -> None:
        pass

    @abstractmethod
    def handle_item_removed(self, index: int, item: T) -> None:
        pass


class ObservableSequence(Sequence[T]):
    """Sequence that maintains a list of observers and dispatches change notifications."""

    def __init__(self) -> None:
        self._observer_list: list[SequenceObserver[T]] = list()

    def add_observer(self, observer: SequenceObserver[T]) -> None:
        if observer not in self._observer_list:
            self._observer_list.append(observer)

    def remove_observer(self, observer: SequenceObserver[T]) -> None:
        try:
            self._observer_list.remove(observer)
        except ValueError:
            pass

    def notify_observers_item_inserted(self, index: int, item: T) -> None:
        for observer in self._observer_list:
            observer.handle_item_inserted(index, item)

    def notify_observers_item_changed(self, index: int, item: T) -> None:
        for observer in self._observer_list:
            observer.handle_item_changed(index, item)

    def notify_observers_item_removed(self, index: int, item: T) -> None:
        for observer in self._observer_list:
            observer.handle_item_removed(index, item)



from pathlib import Path
import logging

from PyQt5.QtCore import QModelIndex, QSortFilterProxyModel, Qt
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QLineEdit,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer, SequenceObserver
from ptychodus.api.parametric import PathParameter

from ...model.globus import GlobusAuthorizer, GlobusSettings, GlobusStatus, GlobusStatusRepository
from ..data import FileDialogFactory
from ..parametric import ParameterViewBuilder, ParameterViewController
from .authorization import GlobusAuthorizationController
from .status import GlobusStatusViewController
from .table_model import GlobusStatusTableModel

logger = logging.getLogger(__name__)


class PathLineEditParameterViewController(ParameterViewController, Observer):
    # TODO: to ptychodus.controller.parametric

    def __init__(self, parameter: PathParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._parameter = parameter
        self._widget = QLineEdit()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self.__sync_model_to_view()
        self._widget.editingFinished.connect(self.__sync_view_to_model)
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def __sync_view_to_model(self) -> None:
        self._parameter.set_value(Path(self._widget.text()))

    def __sync_model_to_view(self) -> None:
        self._widget.setText(str(self._parameter.get_value()))

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.__sync_model_to_view()


class GlobusController(SequenceObserver[GlobusStatus]):
    def __init__(
        self,
        settings: GlobusSettings,
        authorizer: GlobusAuthorizer,
        status_repository: GlobusStatusRepository,
        view: QWidget,
        status_table_view: QTableView,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._view = view
        self._auth_controller = GlobusAuthorizationController(authorizer, view)
        self._compute_data_posix_path_controller = PathLineEditParameterViewController(
            settings.compute_data_posix_path,
            tool_tip='POSIX path on the compute endpoint where data will be stored.',
        )
        self._output_data_posix_path_controller = PathLineEditParameterViewController(
            settings.output_data_posix_path,
            tool_tip='POSIX path on the output data endpoint where data will be stored.',
        )
        self._status_controller = GlobusStatusViewController(
            settings.status_auto_refresh, settings.status_refresh_interval_s, status_repository
        )

        self._status_table_model = GlobusStatusTableModel(status_repository)
        self._status_proxy_model = QSortFilterProxyModel()
        self._status_proxy_model.setSourceModel(self._status_table_model)

        status_table_view.setModel(self._status_proxy_model)
        status_table_view.setSortingEnabled(True)
        status_table_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        status_table_view.clicked.connect(self._handle_table_view_clicked)

        status_repository.add_observer(self)

        view_builder = ParameterViewBuilder(file_dialog_factory)

        input_data_group = 'Input Data'
        view_builder.add_uuid_line_edit(
            settings.input_data_endpoint_id, 'Endpoint ID:', group=input_data_group
        )
        view_builder.add_line_edit(
            settings.input_data_globus_path, 'Globus Path:', group=input_data_group
        )
        view_builder.add_directory_chooser(
            settings.input_data_posix_path, 'POSIX Path:', group=input_data_group
        )

        compute_group = 'Compute'
        view_builder.add_uuid_line_edit(
            settings.compute_endpoint_id, 'Compute Endpoint ID:', group=compute_group
        )
        view_builder.add_uuid_line_edit(
            settings.compute_data_endpoint_id, 'Data Endpoint ID:', group=compute_group
        )
        view_builder.add_line_edit(
            settings.compute_data_globus_path, 'Data Globus Path:', group=compute_group
        )
        view_builder.add_view_controller(
            self._compute_data_posix_path_controller, 'Data POSIX Path:', group=compute_group
        )

        output_data_group = 'Output Data'
        view_builder.add_uuid_line_edit(
            settings.output_data_endpoint_id, 'Endpoint ID:', group=output_data_group
        )
        view_builder.add_line_edit(
            settings.output_data_globus_path, 'Globus Path:', group=output_data_group
        )
        view_builder.add_view_controller(
            self._output_data_posix_path_controller, 'POSIX Path:', group=output_data_group
        )

        view_builder.add_view_controller_to_bottom(self._status_controller)
        contents = view_builder.build_widget()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(contents)
        view.setLayout(layout)

    def _handle_table_view_clicked(self, index: QModelIndex) -> None:
        if index.column() == 5:
            url = index.data(Qt.ItemDataRole.UserRole)
            logger.info(f'Opening URL: "{url.toString()}"')
            QDesktopServices.openUrl(url)

    def run_tasks(self, one_second_counter: int) -> None:
        self._auth_controller.authorize_as_needed()

    def handle_item_inserted(self, index: int, item: GlobusStatus) -> None:
        self._status_table_model.beginInsertRows(QModelIndex(), index, index)
        self._status_table_model.endInsertRows()

    def handle_item_changed(self, index: int, item: GlobusStatus) -> None:
        top_left = self._status_table_model.index(index, 0)
        bottom_right = self._status_table_model.index(index, self._status_table_model.columnCount())
        self._status_table_model.dataChanged.emit(top_left, bottom_right)

    def handle_item_removed(self, index: int, item: GlobusStatus) -> None:
        self._status_table_model.beginRemoveRows(QModelIndex(), index, index)
        self._status_table_model.endRemoveRows()



from collections.abc import Sequence
from pathlib import Path
import logging
import re

from PyQt5.QtCore import Qt, QModelIndex, QSortFilterProxyModel
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QButtonGroup,
    QComboBox,
    QFileSystemModel,
    QFormLayout,
    QHBoxLayout,
    QHeaderView,
    QLineEdit,
    QPushButton,
    QTableView,
    QWidget,
    QWizardPage,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import PathParameter

from ....model.diffraction import DiffractionAPI, DiffractionSettings
from ....view.diffraction import OpenDatasetWizardPage
from ....view.widgets import ExceptionDialog
from ...data import FileDialogFactory
from ...helpers import connect_current_changed_signal

logger = logging.getLogger(__name__)


class OpenDatasetWizardBreadcrumbsViewController(Observer):
    def __init__(self, file_dialog_factory: FileDialogFactory) -> None:
        super().__init__()
        self._file_dialog_factory = file_dialog_factory
        self._widget = QWidget()
        self._path_list: list[Path] = []
        self._button_group = QButtonGroup()
        self._button_group.idClicked.connect(self._handle_id_clicked)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        file_dialog_factory.add_observer(self)

    def _handle_id_clicked(self, button_id: int) -> None:
        path = self._path_list[button_id]
        self._file_dialog_factory.set_open_working_directory(path)

    def _sync_model_to_view(self) -> None:
        path = self._file_dialog_factory.get_open_working_directory().resolve()

        for button_id, existing_path in enumerate(self._path_list):
            if path == existing_path:
                button = self._button_group.button(button_id)
                button.setChecked(True)
                return

        layout = self._widget.layout()

        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                widget = item.widget()

                if widget is None:
                    continue
                elif isinstance(widget, QPushButton):
                    self._button_group.removeButton(widget)
                    widget.deleteLater()

        self._path_list.clear()
        button_list: list[QPushButton] = []

        while True:
            if path.name:
                button = QPushButton(path.name)
                button.setCheckable(True)
                button_list.append(button)
                self._path_list.append(path)
                path = path.parent
            else:
                button = QPushButton(path.anchor)
                button.setCheckable(True)
                button_list.append(button)
                self._path_list.append(Path(path.anchor))
                break

        for button_id, button in reversed(list(enumerate(button_list))):
            self._button_group.addButton(button, button_id)
            layout.addWidget(button)

        if isinstance(layout, QHBoxLayout):
            layout.addStretch()

        button_list[0].setChecked(True)
        self._widget.setLayout(layout)
        self._widget.update()

    def get_widget(self) -> QWidget:
        return self._widget

    def _update(self, observable: Observable) -> None:
        if observable is self._file_dialog_factory:
            self._sync_model_to_view()


class OpenDatasetWizardLocationViewController(Observer):
    def __init__(self, file_path: PathParameter, file_dialog_factory: FileDialogFactory) -> None:
        super().__init__()
        self._file_path = file_path
        self._file_dialog_factory = file_dialog_factory

        self._widget = QLineEdit()
        self._widget.editingFinished.connect(self._handle_editing_finished)

        self._sync_model_to_view()
        file_path.add_observer(self)
        file_dialog_factory.add_observer(self)

    def _handle_editing_finished(self) -> None:
        text = self._widget.text()
        path = Path(text)

        if not path.is_absolute():
            path = self._file_dialog_factory.get_open_working_directory() / text

        path = path.resolve()

        self._file_dialog_factory.set_open_working_directory(path)
        self._file_path.set_value(path)

    def _sync_model_to_view(self) -> None:
        file_path = self._file_path.get_value()

        if file_path.is_file():
            self._widget.setText(file_path.name)
        else:
            self._widget.clear()

    def get_widget(self) -> QWidget:
        return self._widget

    def _update(self, observable: Observable) -> None:
        if observable in (self._file_path, self._file_dialog_factory):
            self._sync_model_to_view()


class OpenDatasetWizardFilePathViewController(Observer):
    def __init__(self, file_path: PathParameter, file_dialog_factory: FileDialogFactory) -> None:
        super().__init__()
        self._file_path = file_path
        self._file_dialog_factory = file_dialog_factory

        self._model = QFileSystemModel()
        self._model.setNameFilterDisables(False)

        self._proxy_model = QSortFilterProxyModel()
        self._proxy_model.setSourceModel(self._model)

        self._widget = QTableView()
        self._widget.setModel(self._proxy_model)
        self._widget.setSortingEnabled(True)
        self._widget.sortByColumn(0, Qt.SortOrder.AscendingOrder)
        self._widget.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.ResizeToContents
        )
        self._widget.verticalHeader().hide()
        self._widget.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self._widget.doubleClicked.connect(self._handle_table_double_clicked)

        connect_current_changed_signal(self._widget, self._handle_current_changed)

        self._sync_model_to_view()
        file_path.add_observer(self)
        file_dialog_factory.add_observer(self)

    def set_name_filters(self, name_filters: Sequence[str]) -> None:
        logger.debug(f'Name Filters: {name_filters}')
        self._model.setNameFilters(name_filters)

    def _handle_table_double_clicked(self, proxy_index: QModelIndex) -> None:
        index = self._proxy_model.mapToSource(proxy_index)
        file_info = self._model.fileInfo(index)

        if file_info.isDir():
            directory = Path(file_info.canonicalFilePath())
            self._file_dialog_factory.set_open_working_directory(directory)

    def _handle_current_changed(self, current: QModelIndex, previous: QModelIndex) -> None:
        index = self._proxy_model.mapToSource(current)
        path = Path(self._model.filePath(index))
        self._file_path.set_value(path)

    def _sync_model_to_view(self) -> None:
        file_path = self._file_path.get_value()
        root_path = self._file_dialog_factory.get_open_working_directory()

        index = self._model.setRootPath(str(root_path))
        proxy_index = self._proxy_model.mapFromSource(index)
        self._widget.setRootIndex(proxy_index)

        if file_path.is_relative_to(root_path):
            index = self._model.index(str(file_path))
            proxy_index = self._proxy_model.mapFromSource(index)
            self._widget.setCurrentIndex(proxy_index)

    def get_widget(self) -> QWidget:
        return self._widget

    def _update(self, observable: Observable) -> None:
        if observable is self._file_path:
            self._sync_model_to_view()
        elif observable is self._file_dialog_factory:
            self._sync_model_to_view()


class OpenDatasetWizardFileTypeViewController(Observable, Observer):
    def __init__(self, api: DiffractionAPI) -> None:
        super().__init__()
        self._file_reader_chooser = api.get_file_reader_chooser()
        self._file_reader_chooser.add_observer(self)
        self._combo_box = QComboBox()

        for plugin in self._file_reader_chooser:
            self._combo_box.addItem(plugin.display_name)

        self._sync_model_to_view()
        self._combo_box.textActivated.connect(self._handle_text_activated)

    def get_name_filters(self) -> Sequence[str]:
        text = self._combo_box.currentText()
        z = re.search(r'\((.+)\)', text)
        return z.group(1).split() if z else []

    def _handle_text_activated(self, text: str) -> None:
        self._file_reader_chooser.set_current_plugin(text)

    def _sync_model_to_view(self) -> None:
        self._combo_box.setCurrentText(self._file_reader_chooser.get_current_plugin().display_name)

    def get_widget(self) -> QWidget:
        return self._combo_box

    def _update(self, observable: Observable) -> None:
        if observable is self._file_reader_chooser:
            self._sync_model_to_view()
            self.notify_observers()


class OpenDatasetWizardFilesViewController(Observer):
    def __init__(
        self,
        settings: DiffractionSettings,
        api: DiffractionAPI,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._api = api
        self._file_dialog_factory = file_dialog_factory

        self._breadcrumbs_view_controller = OpenDatasetWizardBreadcrumbsViewController(
            file_dialog_factory
        )
        self._location_view_controller = OpenDatasetWizardLocationViewController(
            settings.file_path, file_dialog_factory
        )
        self._file_path_view_controller = OpenDatasetWizardFilePathViewController(
            settings.file_path, file_dialog_factory
        )
        self._file_type_view_controller = OpenDatasetWizardFileTypeViewController(api)
        self._file_type_view_controller.add_observer(self)

        layout = QFormLayout()
        layout.addRow(self._breadcrumbs_view_controller.get_widget())
        layout.addRow('Location:', self._location_view_controller.get_widget())
        layout.addRow(self._file_path_view_controller.get_widget())
        layout.addRow('File Type:', self._file_type_view_controller.get_widget())

        self._page = OpenDatasetWizardPage()
        self._page.setTitle('Choose Dataset File(s)')
        self._page.setLayout(layout)

        self._sync_model_to_view()
        settings.file_path.add_observer(self)

    def open_dataset(self) -> None:
        file_reader_chooser = self._api.get_file_reader_chooser()
        file_type = file_reader_chooser.get_current_plugin().simple_name
        file_path = self._settings.file_path.get_value()

        try:
            self._api.open_patterns(file_path, file_type=file_type)
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Open Dataset', err)

    def get_widget(self) -> QWizardPage:
        return self._page

    def _check_if_complete(self) -> None:
        file_path = self._settings.file_path.get_value()
        self._page._set_complete(file_path.is_file())

    def restart(self) -> None:
        self._check_if_complete()

    def _handle_file_type_changed(self) -> None:
        name_filters = self._file_type_view_controller.get_name_filters()
        self._file_path_view_controller.set_name_filters(name_filters)

    def _sync_model_to_view(self) -> None:
        file_path = self._settings.file_path.get_value()

        if file_path.exists():
            self._file_dialog_factory.set_open_working_directory(file_path)

        self._handle_file_type_changed()

    def _update(self, observable: Observable) -> None:
        if observable is self._settings.file_path:
            self._check_if_complete()
        elif observable is self._file_type_view_controller:
            self._handle_file_type_changed()



from decimal import Decimal
import logging

from ptychodus.api.observer import Observable, Observer

from ...model.analysis import ProbePropagator
from ...model.visualization import VisualizationEngine
from ...view.probe import ProbePropagationDialog
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..visualization import (
    VisualizationParametersController,
    VisualizationWidgetController,
)

logger = logging.getLogger(__name__)


class ProbePropagationViewController(Observer):
    def __init__(
        self,
        propagator: ProbePropagator,
        engine: VisualizationEngine,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._propagator = propagator
        self._file_dialog_factory = file_dialog_factory

        self._dialog = ProbePropagationDialog()
        self._dialog.propagate_button.clicked.connect(self._propagate)
        self._dialog.save_button.clicked.connect(self._save_propagated_probe)
        self._dialog.coordinate_slider.valueChanged.connect(self._update_current_coordinate)
        self._dialog.parameters_view.num_steps_spin_box.setRange(1, 999)

        self._xy_visualization_widget_controller = VisualizationWidgetController(
            engine, self._dialog.xy_view, self._dialog.status_bar, file_dialog_factory
        )
        self._zx_visualization_widget_controller = VisualizationWidgetController(
            engine, self._dialog.zx_view, self._dialog.status_bar, file_dialog_factory
        )
        self._visualization_parameters_controller = VisualizationParametersController(
            engine, self._dialog.parameters_view.visualization_parameters_view
        )
        self._zy_visualization_widget_controller = VisualizationWidgetController(
            engine, self._dialog.zy_view, self._dialog.status_bar, file_dialog_factory
        )

        propagator.add_observer(self)
        self._sync_model_to_view()

    def _update_current_coordinate(self, step: int) -> None:
        lerp_value = 0.0

        slider = self._dialog.coordinate_slider
        upper = step - slider.minimum()
        lower = slider.maximum() - slider.minimum()

        if lower > 0:
            alpha = upper / lower
            z0 = self._propagator.get_begin_coordinate_m()
            z1 = self._propagator.get_end_coordinate_m()
            lerp_value = (1 - alpha) * z0 + alpha * z1
        else:
            logger.error('Bad slider range!')

        try:
            xy_projection = self._propagator.get_xy_projection(step)
        except (IndexError, ValueError):
            self._xy_visualization_widget_controller.clear_array()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Update Current Coordinate', err)
        else:
            pixel_geometry = self._propagator.get_pixel_geometry()

            if pixel_geometry is None:
                logger.warning('Missing propagator pixel geometry!')
            else:
                self._xy_visualization_widget_controller.set_array(xy_projection, pixel_geometry)

        # TODO auto-units
        lerp_value *= 1e6
        self._dialog.coordinate_label.setText(f'{lerp_value:.1f} \u00b5m')

    def _propagate(self) -> None:
        view = self._dialog.parameters_view

        try:
            self._propagator.propagate(
                num_steps=view.num_steps_spin_box.value(),
                begin_coordinate_m=float(view.begin_coordinate_widget.get_length_m()),
                end_coordinate_m=float(view.end_coordinate_widget.get_length_m()),
            )
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Propagate Probe', err)

    def launch(self, product_index: int) -> None:
        self._propagator.set_product(product_index)

        try:
            item_name = self._propagator.get_product_name()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Launch', err)
        else:
            self._dialog.setWindowTitle(f'Propagate Probe: {item_name}')
            self._dialog.open()

    def _save_propagated_probe(self) -> None:
        title = 'Save Propagated Probe'
        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._dialog,
            title,
            name_filters=self._propagator.get_save_file_filters(),
            selected_name_filter=self._propagator.get_save_file_filter(),
        )

        if file_path:
            try:
                self._propagator.save_propagated_probe(file_path)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception(title, err)

    def _sync_model_to_view(self) -> None:
        view = self._dialog.parameters_view
        view.begin_coordinate_widget.set_length_m(
            Decimal(repr(self._propagator.get_begin_coordinate_m()))
        )
        view.end_coordinate_widget.set_length_m(
            Decimal(repr(self._propagator.get_end_coordinate_m()))
        )
        view.num_steps_spin_box.setValue(self._propagator.get_num_steps())

        num_steps = self._propagator.get_num_steps()

        if num_steps > 1:
            self._dialog.coordinate_slider.setEnabled(True)
            self._dialog.coordinate_slider.setRange(0, num_steps - 1)
        else:
            self._dialog.coordinate_slider.setEnabled(False)
            self._dialog.coordinate_slider.setRange(0, 1)
            self._dialog.coordinate_slider.setValue(0)

        self._update_current_coordinate(self._dialog.coordinate_slider.value())
        pixel_geometry = self._propagator.get_pixel_geometry()

        if pixel_geometry is None:
            logger.warning('Missing propagator pixel geometry!')
            return

        try:
            # vvv TODO display correct pixel geometry for projections vvv
            self._zx_visualization_widget_controller.set_array(
                self._propagator.get_zx_projection(), pixel_geometry
            )
            self._zy_visualization_widget_controller.set_array(
                self._propagator.get_zy_projection(), pixel_geometry
            )
        except ValueError:
            self._zx_visualization_widget_controller.clear_array()
            self._zy_visualization_widget_controller.clear_array()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Update Views', err)

    def _update(self, observable: Observable) -> None:
        if observable is self._propagator:
            self._sync_model_to_view()



from __future__ import annotations
from decimal import Decimal
import logging

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import QHBoxLayout, QLineEdit, QWidget

logger = logging.getLogger(__name__)


class DecimalLineEdit(QWidget):
    value_changed = pyqtSignal(Decimal)

    def __init__(self, parent: QWidget | None) -> None:
        super().__init__(parent)
        self._validator = QDoubleValidator()
        self._line_edit = QLineEdit()
        self._value = Decimal()
        self._minimum: Decimal | None = None
        self._maximum: Decimal | None = None

    @classmethod
    def create_instance(
        cls, *, is_signed: bool = False, parent: QWidget | None = None
    ) -> DecimalLineEdit:
        widget = cls(parent)

        widget._line_edit.setValidator(widget._validator)
        widget._line_edit.editingFinished.connect(widget._set_value_from_line_edit)
        widget._set_value_to_line_edit_and_emit_value_changed()

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget._line_edit)
        widget.setLayout(layout)

        if not is_signed:
            widget._validator.setBottom(0.0)

        return widget

    def is_read_only(self) -> bool:
        return self._line_edit.isReadOnly()

    def set_read_only(self, enable: bool) -> None:
        self._line_edit.setReadOnly(enable)

    def get_value(self) -> Decimal:
        if self._minimum is not None and self._value < self._minimum:
            return self._minimum

        if self._maximum is not None and self._value > self._maximum:
            return self._maximum

        return self._value

    def set_value(self, value: Decimal) -> None:
        if value != self._value:
            self._value = value
            self._set_value_to_line_edit_and_emit_value_changed()

    def get_minimum(self) -> Decimal | None:
        return self._minimum

    def set_minimum(self, value: Decimal) -> None:
        value_before = self.get_value()
        self._minimum = value
        value_after = self.get_value()

        if value_before != value_after:
            self._set_value_to_line_edit_and_emit_value_changed()

    def get_maximum(self) -> Decimal | None:
        return self._maximum

    def set_maximum(self, value: Decimal) -> None:
        value_before = self.get_value()
        self._maximum = value
        value_after = self.get_value()

        if value_before != value_after:
            self._set_value_to_line_edit_and_emit_value_changed()

    def _set_value_from_line_edit(self) -> None:
        decimal_text = self._line_edit.text()

        try:
            self._value = Decimal(decimal_text)
        except ValueError:
            logger.error(f'Failed to parse value "{decimal_text}"')
        else:
            self._emit_value_changed()

    def _set_value_to_line_edit_and_emit_value_changed(self) -> None:
        self._line_edit.setText(str(self.get_value()))
        self._emit_value_changed()

    def _emit_value_changed(self) -> None:
        self.value_changed.emit(self._value)



from collections import defaultdict
from pathlib import Path
from typing import Final
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    DiffractionArray,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

from .h5_diffraction_file import H5DiffractionPatternArray

logger = logging.getLogger(__name__)


class APS12IDDiffractionFileReader(DiffractionFileReader):
    DATA_PATH: Final[str] = '/entry/data/data'

    def read(self, file_path: Path) -> DiffractionDataset:
        stem_parts = file_path.stem.split('_')
        stem_prefix = '_'.join(stem_parts[:-2])
        logger.debug(f'{stem_prefix=}')
        scan_num = int(stem_parts[-3])
        logger.debug(f'{scan_num=}')

        lines: set[int] = set()
        points: set[int] = set()
        points_per_line: dict[int, set[int]] = defaultdict(set[int])
        file_dict: dict[tuple[int, int], Path] = dict()

        for p in file_path.parent.glob(f'{stem_prefix}_*{file_path.suffix}'):
            stem_parts = p.stem.split('_')
            line = int(stem_parts[-2])
            point = int(stem_parts[-1])

            lines.add(line)
            points.add(point)
            points_per_line[line].add(point)
            file_dict[line, point] = p

        logger.debug(f'{lines=}')
        lines_min = min(lines)
        lines_max = max(lines)
        lines_num = lines_max - lines_min + 1
        logger.debug(f'{points=}')
        points_min = min(points)
        points_max = max(points)
        points_num = points_max - points_min + 1

        for line, line_points in points_per_line.items():
            missing_points = points - line_points

            if missing_points:
                logger.warning(f'Line {line} is missing points {missing_points}')

        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Line', 'Point'])
        array_list: list[DiffractionArray] = list()

        with h5py.File(file_path, 'r') as h5_file:
            try:
                h5_data = h5_file[self.DATA_PATH]
            except KeyError:
                logger.warning(f'File "{file_path}" is not an APS 12-ID data file.')
                return SimpleDiffractionDataset.create_null(file_path)
            else:
                if not isinstance(h5_data, h5py.Dataset):
                    logger.warning(
                        f'Data path "{self.DATA_PATH}" in "{file_path}" is not a dataset.'
                    )
                    return SimpleDiffractionDataset.create_null(file_path)

                detector_height, detector_width = h5_data.shape

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[1] * lines_num * points_num,
                    pattern_dtype=h5_data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    file_path=file_path,
                )

        for (line, point), fp in file_dict.items():
            index = (point - points_min) + (line - lines_min) * points_num
            indexes = numpy.array([index])
            array = H5DiffractionPatternArray(fp.stem, indexes, fp, self.DATA_PATH)
            contents_tree.create_child([array.get_label(), 'HDF5', str(line), str(point)])
            array_list.append(array)

        return SimpleDiffractionDataset(metadata, contents_tree, array_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        APS12IDDiffractionFileReader(),
        simple_name='APS_PtychoSAXS',
        display_name='APS 12-ID PtychoSAXS Files (*.h5 *.hdf5)',
    )



from typing import Any
import logging

import numpy

from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QObject

from ...model.product.probe import ZernikeProbeBuilder

logger = logging.getLogger(__name__)


class ZernikeTableModel(QAbstractTableModel):
    def __init__(self, builder: ZernikeProbeBuilder, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._builder = builder
        self._header = [
            'Radial Degree',
            'Angular Frequency',
            'Amplitude',
            'Phase [tr]',
        ]

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid() and index.column() in (2, 3):
            value |= Qt.ItemFlag.ItemIsEditable

        return value

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None

        try:
            poly = self._builder.get_polynomial(index.row())
            coef = self._builder.get_coefficient(index.row())
        except IndexError as err:
            logger.exception(err)
            return None

        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                return poly.radial_degree
            elif index.column() == 1:
                return poly.angular_frequency
            elif index.column() == 2:
                return f'{numpy.absolute(coef):.6g}'
            elif index.column() == 3:
                return f'{numpy.angle(coef):.6g}'

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if not index.isValid():
            return False

        if role == Qt.ItemDataRole.EditRole:
            if index.column() == 2:
                try:
                    amplitude = float(value)
                except ValueError:
                    return False

                try:
                    coef = self._builder.get_coefficient(index.row())
                except IndexError:
                    return False

                try:
                    complex_value = amplitude * coef / numpy.absolute(coef)
                except ZeroDivisionError:
                    complex_value = amplitude + 0j

                self._builder.set_coefficient(index.row(), complex_value)
                return True
            elif index.column() == 3:
                try:
                    phase = float(value)
                except ValueError:
                    return False

                try:
                    coef = self._builder.get_coefficient(index.row())
                except IndexError:
                    return False

                complex_value = numpy.absolute(coef) * numpy.exp(2j * numpy.pi * phase)
                self._builder.set_coefficient(index.row(), complex_value)
                return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._builder)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)



from __future__ import annotations
from collections.abc import Sequence
import logging

from scipy.interpolate import griddata
import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider

from ...diffraction import DiffractionAPI
from .builder import ObjectBuilder
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class STXMObjectBuilder(ObjectBuilder):
    def __init__(
        self,
        settings: ObjectSettings,
        diffraction_api: DiffractionAPI,
    ) -> None:
        super().__init__(settings, 'stxm')
        self._settings = settings
        self._diffraction_api = diffraction_api

    def copy(self) -> STXMObjectBuilder:
        builder = STXMObjectBuilder(self._settings, self._diffraction_api)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        geometry = geometry_provider.get_object_geometry()
        coordinates_px: list[float] = list()
        values: list[float] = list()

        assembled_data = self._diffraction_api.get_assembled_data()
        assembled_indexes = assembled_data.get_indexes().tolist()
        assembled_pattern_counts = assembled_data.get_pattern_counts().tolist()
        pattern_counts_lut = dict(zip(assembled_indexes, assembled_pattern_counts))

        for scan_point in geometry_provider.get_probe_positions():
            try:
                value = pattern_counts_lut[scan_point.index]
            except KeyError:
                logger.debug(f'Skipping missing scan point index={scan_point.index}!')
            else:
                object_point = geometry.map_coordinates_probe_to_object(scan_point)
                coordinates_px.append(object_point.coordinate_y_px)
                coordinates_px.append(object_point.coordinate_x_px)
                values.append(value)

        points = numpy.reshape(coordinates_px, (-1, 2))
        YY, XX = numpy.mgrid[: geometry.height_px, : geometry.width_px]  # noqa: N806
        query_points = numpy.transpose((YY.flat, XX.flat))
        intensity = griddata(points, values, query_points, method='linear', fill_value=0.0).reshape(
            XX.shape
        )

        return self._create_object(
            array=numpy.sqrt(intensity[numpy.newaxis, :, :]).astype('complex'),
            layer_spacing_m=layer_spacing_m,
            pixel_geometry=geometry.get_pixel_geometry(),
            center=geometry.get_center(),
        )



from abc import ABC, abstractmethod
from typing import Callable, Final, TypeAlias
import logging
import queue
import threading

logger = logging.getLogger(__name__)

ForegroundTask: TypeAlias = Callable[[], None]


class ForegroundTaskManager(ABC):
    @abstractmethod
    def put_foreground_task(self, task: ForegroundTask) -> None:
        pass

    @property
    @abstractmethod
    def foreground_queue_size(self) -> int:
        pass


BackgroundTask: TypeAlias = Callable[[], ForegroundTask | None]


class BackgroundTaskManager(ABC):
    @abstractmethod
    def put_background_task(self, task: BackgroundTask) -> None:
        pass

    @property
    @abstractmethod
    def background_queue_size(self) -> int:
        pass


class TaskManager(BackgroundTaskManager, ForegroundTaskManager):
    WAIT_TIME_S: Final[float] = 1.0

    def __init__(self) -> None:
        super().__init__()
        self._background_queue: queue.Queue[BackgroundTask] = queue.Queue()
        self._foreground_queue: queue.Queue[ForegroundTask] = queue.Queue()
        self._stop_event = threading.Event()
        self._worker: threading.Thread | None = None

    @property
    def is_stopping(self) -> bool:
        return self._stop_event.is_set()

    def put_background_task(self, task: BackgroundTask) -> None:
        self._background_queue.put(task)

    @property
    def background_queue_size(self) -> int:
        return self._background_queue.qsize()

    def _run_background_tasks(self) -> None:
        while not self._stop_event.is_set():
            try:
                background_task = self._background_queue.get(block=True, timeout=self.WAIT_TIME_S)
            except queue.Empty:
                continue

            try:
                foreground_task = background_task()
            except Exception:
                logger.exception(f'Background task exception during {background_task}!')
            else:
                if foreground_task is not None:
                    self._foreground_queue.put(foreground_task)
            finally:
                self._background_queue.task_done()

    def put_foreground_task(self, task: ForegroundTask) -> None:
        self._foreground_queue.put(task)

    @property
    def foreground_queue_size(self) -> int:
        return self._foreground_queue.qsize()

    def run_foreground_tasks(self) -> None:
        while True:
            try:
                task = self._foreground_queue.get(block=False)
            except queue.Empty:
                break

            try:
                task()
            except Exception:
                logger.exception(f'Foreground task exception during {task}!')
            finally:
                self._foreground_queue.task_done()

    def start(self) -> None:
        if self._worker is None:
            logger.info('Starting task manager...')
            self._stop_event.clear()
            self._worker = threading.Thread(target=self._run_background_tasks)
            self._worker.start()
            logger.info('Task manager started.')
        else:
            logger.warning('Worker already started!')

    def _stop(self) -> None:
        if self._worker is None:
            logger.warning('Worker is None!')
        else:
            logger.info('Stopping task manager...')
            self._stop_event.set()
            self._worker.join()
            self._worker = None
            logger.info('Task manager stopped.')

    def stop(self, *, await_finish: bool) -> None:
        if self._stop_event.is_set():
            logger.info('Task manager already stopped.')
        else:
            if await_finish:
                logger.info('Finishing tasks...')
                self._background_queue.join()
                logger.info('Tasks finished.')

            self._stop()



from __future__ import annotations
from datetime import datetime, timezone
from importlib.metadata import version
from pprint import pformat
import json
import logging
import queue
import threading

import fair_research_login
import gladier

from ptychodus.api.observer import Observable, Observer

from ..task_manager import TaskManager
from ._gladier_client import PtychodusClient
from .authorizer import AuthorizeWithGlobus, GlobusAuthorizer
from .client import GlobusClient, GlobusJob
from .settings import GlobusSettings
from .status import GlobusStatus, GlobusStatusRepository, UpdateGlobusStatusRepository

__all__ = ['RealGlobusClient']

logger = logging.getLogger(__name__)


class CustomCodeHandler(fair_research_login.CodeHandler):
    def __init__(
        self, task_manager: TaskManager, stop_event: threading.Event, authorizer: GlobusAuthorizer
    ) -> None:
        super().__init__()
        self._task_manager = task_manager
        self._stop_event = stop_event
        self._authorizer = authorizer

        self._user_provided_code = threading.Event()
        self._code_lock = threading.Lock()
        self._code = ''

        self.set_browser_enabled(False)

    def authenticate(self, url: str) -> str:
        self._user_provided_code.clear()

        with self._code_lock:
            self._code = ''

        task = AuthorizeWithGlobus(self._authorizer, url)
        self._task_manager.put_foreground_task(task)

        while not self._stop_event.is_set():
            if self._user_provided_code.wait(timeout=1.0):
                break

        return self.get_code()

    def set_code_from_user(self, code: str) -> None:
        with self._code_lock:
            self._code = code

        self._user_provided_code.set()

    def get_code(self) -> str:
        with self._code_lock:
            return self._code


class RunGlobusFlow:
    def __init__(self, gladier_client: gladier.GladierBaseClient, job: GlobusJob) -> None:
        self._gladier_client = gladier_client
        self._job = job

    def __call__(self) -> None:
        try:
            response = self._gladier_client.run_flow(
                flow_input={'input': self._job.flow_input},
                label=self._job.label,
                tags=self._job.tags,
            )
        except Exception:
            logger.exception('Error running flow!')
        else:
            logger.info(f'Run Flow Response: {json.dumps(response, indent=4)}')


class RealGlobusClient(GlobusClient, Observer):
    def __init__(
        self,
        task_manager: TaskManager,
        settings: GlobusSettings,
        authorizer: GlobusAuthorizer,
        status_repository: GlobusStatusRepository,
    ) -> None:
        super().__init__()
        self._task_manager = task_manager
        self._settings = settings
        self._authorizer = authorizer
        self._status_repository = status_repository

        logger.info('\tGlobus SDK ' + version('globus-sdk'))
        logger.info('\tFair Research Login ' + version('fair-research-login'))
        logger.info('\tGladier ' + version('gladier'))

        self._stop_event = threading.Event()
        self._job_queue: queue.Queue[GlobusJob] = queue.Queue()
        self._worker: threading.Thread | None = None
        self.__gladier_client: gladier.GladierBaseClient | None = None

        self._code_handler = CustomCodeHandler(task_manager, self._stop_event, authorizer)
        authorizer.add_observer(self)

        self._status_date_time = datetime.min
        self._status_lock = threading.Lock()
        self._status_auto_refresh = False
        self._status_refresh_interval_s = 0

        self._sync_status_refresh_interval()
        settings.status_auto_refresh.add_observer(self)
        settings.status_refresh_interval_s.add_observer(self)

    @property
    def _gladier_client(self) -> gladier.GladierBaseClient:
        if self.__gladier_client is None:
            self.__gladier_client = PtychodusClient.create_client([self._code_handler])

        return self.__gladier_client

    @property
    def is_supported(self) -> bool:
        return True

    def start(self) -> None:
        if self._worker is None:
            logger.debug('Starting Globus thread...')
            self._stop_event.clear()
            self._worker = threading.Thread(target=self._run_tasks)
            self._worker.start()
            logger.debug('Globus thread started.')
        else:
            logger.warning('Worker already started!')

    def stop(self) -> None:
        if self._stop_event.is_set():
            logger.debug('Globus thread already stopped.')
        else:
            logger.debug('Finishing tasks...')
            self._job_queue.join()
            logger.debug('Tasks finished.')

            if self._worker is None:
                logger.warning('Worker is None!')
            else:
                logger.debug('Stopping Globus thread...')
                self._stop_event.set()
                self._worker.join()
                self._worker = None
                logger.debug('Globus thread stopped.')

    def run_flow(self, job: GlobusJob) -> None:
        self._job_queue.put(job)

    def _get_current_action(self, run_id: str) -> str:
        status = self._gladier_client.get_status(run_id)
        action = status.get('state_name')

        if not action:
            try:
                det = status['details']
            except Exception:
                logger.exception('Unexpected flow status!')
                logger.error(pformat(status))
            else:
                if det.get('details') and det['details'].get('state_name'):
                    action = det['details']['state_name']
                elif det.get('details') and det['details'].get('output'):
                    action = list(det['details']['output'].keys())[0]
                elif det.get('action_statuses'):
                    action = det['action_statuses'][0].get('state_name')
                elif det.get('code') == 'FlowStarting':
                    pass

        return action

    def _refresh_status(self) -> None:
        status_list: list[GlobusStatus] = list()

        logger.debug('Refreshing status.')

        flows_manager = self._gladier_client.flows_manager
        flow_id = flows_manager.get_flow_id()
        flows_client = flows_manager.flows_client
        response = flows_client.list_runs(filter_flow_id=flow_id)
        run_dict_list = response['runs']

        while response['has_next_page']:
            response = flows_client.list_runs(filter_flow_id=flow_id, marker=response['marker'])
            run_dict_list.extend(response['runs'])

        self._status_date_time = datetime.now(timezone.utc)

        for run_dict in run_dict_list:
            run_id = run_dict.get('run_id', '')
            action = self._get_current_action(run_id)
            start_time_str = run_dict.get('start_time', '')

            try:
                start_time = datetime.fromisoformat(start_time_str)
            except ValueError:
                logger.warning(f'Failed to parse startTime "{start_time_str}"!')
                start_time = datetime.min

            completion_time_str = run_dict.get('completion_time', '')

            try:
                completion_time = datetime.fromisoformat(completion_time_str)
            except ValueError:
                completion_time = None

            status = GlobusStatus(
                label=run_dict.get('label', ''),
                start_time=start_time,
                completion_time=completion_time,
                status=run_dict.get('status', ''),
                action=action,
                run_id=run_id,
            )
            status_list.append(status)

        status_update = UpdateGlobusStatusRepository(self._status_repository, status_list)
        self._task_manager.put_foreground_task(status_update)

    def _run_tasks(self) -> None:
        while not self._stop_event.is_set():
            with self._status_lock:
                if self._status_auto_refresh:
                    status_time_delta = datetime.now(timezone.utc) - self._status_date_time
                    status_age_s = status_time_delta.total_seconds()

                    if status_age_s >= self._status_refresh_interval_s:
                        self._status_repository.refresh_status()

            if self._status_repository.needs_status_refresh():
                self._status_repository._status_refreshed()
                self._refresh_status()

            try:
                job = self._job_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass
            else:
                task = RunGlobusFlow(self._gladier_client, job)
                self._task_manager.put_background_task(task)
                self._job_queue.task_done()

    def _sync_status_refresh_interval(self) -> None:
        with self._status_lock:
            self._status_auto_refresh = self._settings.status_auto_refresh.get_value()
            self._status_refresh_interval_s = self._settings.status_refresh_interval_s.get_value()

    def _update(self, observable: Observable) -> None:
        if observable is self._authorizer:
            if self._authorizer.has_authorize_code:
                self._code_handler.set_code_from_user(self._authorizer.get_authorize_code())
        elif observable in (
            self._settings.status_auto_refresh,
            self._settings.status_refresh_interval_s,
        ):
            self._sync_status_refresh_interval()



from __future__ import annotations
from collections.abc import Iterator
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.visualization import (
    NumberArrayType,
    RealArrayType,
    VisualizationProduct,
)

from .color_axis import ColorAxis
from .colormap import ColormapParameter
from .components import DataArrayComponent
from .renderer import Renderer
from .transformation import ScalarTransformationParameter


class ColormapRenderer(Renderer):
    def __init__(
        self,
        component: DataArrayComponent,
        transformation: ScalarTransformationParameter,
        color_axis: ColorAxis,
        colormap: ColormapParameter,
    ) -> None:
        super().__init__(component.name)
        self._component = component
        self._transformation = transformation
        self._add_parameter('transformation', transformation)
        self._color_axis = color_axis
        self._add_group('color_axis', color_axis, observe=True)
        self._colormap = colormap
        self._add_parameter('colormap', colormap)

    def variants(self) -> Iterator[str]:
        return self._colormap.choices()

    def get_variant(self) -> str:
        return self._colormap.get_value()

    def set_variant(self, variant: str) -> None:
        self._colormap.set_value(variant)

    def is_cyclic(self) -> bool:
        return self._component.is_cyclic

    def _colorize(self, values_transformed: RealArrayType) -> RealArrayType:
        vrange = self._color_axis.get_range()
        norm = Normalize(vmin=vrange.lower, vmax=vrange.upper, clip=False)
        cmap = self._colormap.get_plugin()
        scalar_mappable = ScalarMappable(norm, cmap)
        return scalar_mappable.to_rgba(values_transformed)

    def colorize(self, array: NumberArrayType) -> RealArrayType:
        values = self._component.calculate(array)
        values_transformed = self._transformation.transform(values)
        return self._colorize(values_transformed)

    def render(
        self,
        array: NumberArrayType,
        pixel_geometry: PixelGeometry,
        *,
        autoscale_color_axis: bool,
    ) -> VisualizationProduct:
        values = self._component.calculate(array)
        values_transformed = self._transformation.transform(values)

        if autoscale_color_axis:
            self._color_axis.set_to_data_range(values_transformed)

        rgba = self._colorize(values_transformed)

        return VisualizationProduct(
            value_label=self._transformation.decorate_text(self._component.name),
            values=array,
            rgba=rgba,
            pixel_geometry=pixel_geometry,
        )



from .angle_widget import AngleWidget
from .group_box import BottomTitledGroupBox, GroupBoxWithPresets
from .combo_box_item_delegate import ComboBoxItemDelegate
from .decimal_line_edit import DecimalLineEdit
from .decimal_slider import DecimalSlider
from .exception_dialog import ExceptionDialog
from .length_widget import LengthWidget
from .progress_bar_item_delegate import ProgressBarItemDelegate
from .uuid_line_edit import UUIDLineEdit

__all__ = [
    'AngleWidget',
    'BottomTitledGroupBox',
    'ComboBoxItemDelegate',
    'DecimalLineEdit',
    'DecimalSlider',
    'ExceptionDialog',
    'GroupBoxWithPresets',
    'LengthWidget',
    'ProgressBarItemDelegate',
    'UUIDLineEdit',
]



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure

from .image import ImageView
from .visualization import VisualizationParametersView, VisualizationWidget


class FourierRingCorrelationDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.product1_label = QLabel('Product 1:')
        self.product1_combo_box = QComboBox()
        self.product2_label = QLabel('Product 2:')
        self.product2_combo_box = QComboBox()
        self.figure = Figure()
        self.figure_canvas = FigureCanvasQTAgg(self.figure)
        self.navigation_toolbar = NavigationToolbar(self.figure_canvas, self)
        self.axes = self.figure.add_subplot(111)

        parameters_layout = QGridLayout()
        parameters_layout.addWidget(self.product1_label, 0, 0)
        parameters_layout.addWidget(self.product1_combo_box, 0, 1)
        parameters_layout.addWidget(self.product2_label, 0, 2)
        parameters_layout.addWidget(self.product2_combo_box, 0, 3)
        parameters_layout.setColumnStretch(1, 1)
        parameters_layout.setColumnStretch(3, 1)

        layout = QVBoxLayout()
        layout.addWidget(self.navigation_toolbar)
        layout.addWidget(self.figure_canvas)
        layout.addLayout(parameters_layout)
        self.setLayout(layout)


class FourierAnalysisDialog(QDialog):
    @staticmethod
    def _box_widget(title: str, widget: QWidget) -> QWidget:
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)

        box = QGroupBox(title)
        box.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        box.setLayout(layout)

        return box

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.real_space_view = ImageView(add_fourier_tool=True)
        self.reciprocal_space_view = ImageView()
        self.status_bar = QStatusBar()

        real_space_box = self._box_widget('Real Space', self.real_space_view)
        reciprocal_space_box = self._box_widget('Reciprocal Space', self.reciprocal_space_view)

        contents_layout = QHBoxLayout()
        contents_layout.addWidget(real_space_box)
        contents_layout.addWidget(reciprocal_space_box)

        layout = QVBoxLayout()
        layout.addLayout(contents_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)


class XMCDParametersView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Parameters', parent)

        self.polarization_group_box = QGroupBox('Polarization')
        self.lcirc_combo_box = QComboBox()
        self.rcirc_combo_box = QComboBox()
        self.save_button = QPushButton('Save')
        self.visualization_parameters_view = VisualizationParametersView()

        polarization_layout = QFormLayout()
        polarization_layout.addRow('Left Circular:', self.lcirc_combo_box)
        polarization_layout.addRow('Right Circular:', self.rcirc_combo_box)
        polarization_layout.addRow(self.save_button)
        self.polarization_group_box.setLayout(polarization_layout)

        layout = QVBoxLayout()
        layout.addWidget(self.polarization_group_box)
        layout.addWidget(self.visualization_parameters_view)
        layout.addStretch()
        self.setLayout(layout)


class XMCDDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.difference_widget = VisualizationWidget('Difference')
        self.ratio_widget = VisualizationWidget('Ratio')
        self.sum_widget = VisualizationWidget('Sum')
        self.parameters_view = XMCDParametersView()
        self.status_bar = QStatusBar()

        contents_layout = QGridLayout()
        contents_layout.addWidget(self.difference_widget, 0, 0)
        contents_layout.addWidget(self.ratio_widget, 0, 1)
        contents_layout.addWidget(self.sum_widget, 1, 0)
        contents_layout.addWidget(self.parameters_view, 1, 1)

        layout = QVBoxLayout()
        layout.addLayout(contents_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)



"""Probe (illumination function) data structures and file I/O plugin interfaces."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import overload

import numpy

from .common import ComplexArrayType, RealArrayType
from .geometry import PixelGeometry
from .propagator import intensity


@dataclass(frozen=True)
class FresnelZonePlate:
    """Physical parameters of a Fresnel zone plate optic."""

    zone_plate_diameter_m: float
    outermost_zone_width_m: float
    central_beamstop_diameter_m: float

    def get_focal_length_m(self, central_wavelength_m: float) -> float:
        return self.zone_plate_diameter_m * self.outermost_zone_width_m / central_wavelength_m


@dataclass(frozen=True)
class ProbeGeometry:
    """Pixel dimensions and physical size of the probe array."""

    width_px: int
    height_px: int
    pixel_width_m: float
    pixel_height_m: float

    @property
    def width_m(self) -> float:
        return self.width_px * self.pixel_width_m

    @property
    def height_m(self) -> float:
        return self.height_px * self.pixel_height_m

    def get_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self.pixel_width_m,
            height_m=self.pixel_height_m,
        )


class ProbeGeometryProvider(ABC):
    """Abstract source of detector and probe geometry."""

    @property
    @abstractmethod
    def detector_distance_m(self) -> float:
        pass

    @property
    @abstractmethod
    def probe_photon_count(self) -> float:
        pass

    @property
    @abstractmethod
    def probe_wavelength_m(self) -> float:
        pass

    @property
    @abstractmethod
    def probe_power_W(self) -> float:  # noqa: N802
        pass

    @property
    @abstractmethod
    def num_scan_points(self) -> int:
        pass

    @abstractmethod
    def get_detector_pixel_geometry(self) -> PixelGeometry:
        pass

    @abstractmethod
    def get_probe_geometry(self) -> ProbeGeometry:
        pass


class Probe:
    """Probe (illumination function) stored as a (modes, height, width) complex array."""

    def __init__(
        self,
        array: ComplexArrayType,
        pixel_geometry: PixelGeometry,
    ) -> None:
        self._array = array
        self._pixel_geometry = pixel_geometry

        if array.ndim != 3:
            raise ValueError('Probe must be a 3-dimensional ndarray.')

        power = numpy.sum(intensity(array), axis=(-2, -1))
        powersum = numpy.sum(power)

        if powersum > 0.0:
            power /= powersum

        self._mode_relative_power = power.tolist()

    def copy(self) -> Probe:
        return Probe(
            array=self._array.copy(),
            pixel_geometry=self._pixel_geometry.copy(),
        )

    def get_array(self) -> ComplexArrayType:
        return self._array

    def get_pixel_geometry(self) -> PixelGeometry:
        return self._pixel_geometry

    @property
    def dtype(self) -> numpy.dtype:
        return self._array.dtype

    @property
    def width_px(self) -> int:
        return self._array.shape[-1]

    @property
    def height_px(self) -> int:
        return self._array.shape[-2]

    @property
    def num_incoherent_modes(self) -> int:
        return self._array.shape[-3]

    def get_incoherent_mode(self, number: int) -> ComplexArrayType:
        return self._array[number, :, :]

    def get_incoherent_modes_flattened(self) -> ComplexArrayType:
        return self._array.transpose((1, 0, 2)).reshape(self.height_px, -1)

    def get_incoherent_mode_relative_power(self, number: int) -> float:
        return self._mode_relative_power[number]

    def get_coherence(self) -> float:
        return numpy.sqrt(numpy.sum(numpy.square(self._mode_relative_power)))

    def get_intensity(self) -> RealArrayType:
        return numpy.sum(intensity(self._array), axis=-3)


class ProbeSequence(Sequence[Probe]):
    """Position-dependent probe ensemble stored as a (coherent, incoherent, height, width) array.

    Supports optional OPR (orthogonal probe relaxation) weights for per-position probe variation.
    """

    def __init__(
        self,
        array: ComplexArrayType | None,
        opr_weights: RealArrayType | None,
        pixel_geometry: PixelGeometry | None,
    ) -> None:
        if array is None:
            self._array: ComplexArrayType = numpy.zeros((1, 1, 0, 0), dtype=complex)
        elif numpy.iscomplexobj(array):
            match array.ndim:
                case 2:
                    self._array = array[numpy.newaxis, numpy.newaxis, ...]
                case 3:
                    self._array = array[numpy.newaxis, ...]
                case 4:
                    self._array = array
                case _:
                    raise ValueError('Probe must be 2-, 3-, or 4-dimensional ndarray.')
        else:
            raise TypeError('Probe must be a complex-valued ndarray')

        if opr_weights is None:
            self._opr_weights = None
        elif numpy.issubdtype(opr_weights.dtype, numpy.floating):
            if opr_weights.ndim == 2:
                num_weights_actual = opr_weights.shape[1]
                num_weights_expected = self._array.shape[0]

                if num_weights_actual == num_weights_expected:
                    self._opr_weights = opr_weights
                else:
                    raise ValueError(
                        (
                            'inconsistent number of opr weights!'
                            f' actual={num_weights_actual}'
                            f' expected={num_weights_expected}'
                        )
                    )
            else:
                raise ValueError('opr_weights must be 2-dimensional ndarray')
        else:
            raise TypeError('opr_weights must be a floating-point ndarray')

        self._pixel_geometry = pixel_geometry

    def copy(self) -> ProbeSequence:
        return ProbeSequence(
            self._array.copy(),
            None if self._opr_weights is None else self._opr_weights.copy(),
            None if self._pixel_geometry is None else self._pixel_geometry.copy(),
        )

    def get_array(self) -> ComplexArrayType:
        return self._array

    def get_opr_weights(self) -> RealArrayType:
        if self._opr_weights is None:
            raise ValueError('Missing opr_weights!')

        return self._opr_weights

    def get_pixel_geometry(self) -> PixelGeometry:
        if self._pixel_geometry is None:
            raise ValueError('Missing probe pixel geometry!')

        return self._pixel_geometry

    @property
    def dtype(self) -> numpy.dtype:
        return self._array.dtype

    @property
    def nbytes(self) -> int:
        sz = self._array.nbytes

        if self._opr_weights is not None:
            sz += self._opr_weights.nbytes

        return sz

    @property
    def num_coherent_modes(self) -> int:
        return self._array.shape[0]

    @property
    def num_incoherent_modes(self) -> int:
        return self._array.shape[1]

    @property
    def height_px(self) -> int:
        return self._array.shape[2]

    @property
    def width_px(self) -> int:
        return self._array.shape[3]

    @overload
    def __getitem__(self, index: int) -> Probe: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Probe]: ...

    def __getitem__(self, index: int | slice) -> Probe | Sequence[Probe]:
        if isinstance(index, slice):
            return [self[idx] for idx in range(index.start, index.stop, index.step)]

        array = self._array[0, :, :, :].copy()

        if self._opr_weights is not None:
            array[0, :, :] = numpy.tensordot(
                self._opr_weights[index, :], self._array[:, 0, :, :], axes=1
            )

        return Probe(array, self.get_pixel_geometry())

    def get_probe_no_opr(self) -> Probe:
        array = self._array[0, :, :, :].copy()
        return Probe(array, self.get_pixel_geometry())

    def get_geometry(self) -> ProbeGeometry:
        pixel_geometry = self.get_pixel_geometry()

        return ProbeGeometry(
            width_px=self.width_px,
            height_px=self.height_px,
            pixel_width_m=pixel_geometry.width_m,
            pixel_height_m=pixel_geometry.height_m,
        )

    def __len__(self) -> int:
        return 1 if self._opr_weights is None else self._opr_weights.shape[0]

    def __repr__(self) -> str:
        return f'{self._array.dtype}{self._array.shape}'


class ProbeFileReader(ABC):
    """Plugin interface for reading probe sequences."""

    @abstractmethod
    def read(self, file_path: Path) -> ProbeSequence:
        """Read a probe sequence from file."""
        pass


class ProbeFileWriter(ABC):
    """Plugin interface for writing probe sequences."""

    @abstractmethod
    def write(self, file_path: Path, probes: ProbeSequence) -> None:
        """Write a probe sequence to file."""
        pass



"""Workflow API interfaces for orchestrating ptychography data processing pipelines."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping
from pathlib import Path
from typing import Any

from ptychodus.api.diffraction import CropCenter
from ptychodus.api.geometry import AffineTransform, ImageExtent
from ptychodus.api.product import Product
from ptychodus.api.reconstructor import AssembledDiffractionData, ReconstructInput
from ptychodus.api.settings import PathPrefixChange


class WorkflowDiffractionAPI(ABC):
    """Act on an assembled diffraction dataset within a workflow."""

    @abstractmethod
    def get_assembled_data(self) -> AssembledDiffractionData:
        """Return the assembled diffraction data."""
        pass

    @abstractmethod
    def save_assembled_data(self, file_path: Path) -> None:
        """Save the assembled diffraction data to file."""
        pass


class WorkflowProductAPI(ABC):
    """Act on a data product within a workflow."""

    @abstractmethod
    def get_product_index(self) -> int:
        """Return the registry index of this product."""
        pass

    @abstractmethod
    def get_product(self) -> Product:
        """Return the current product."""
        pass

    @abstractmethod
    def rename_product(self, new_name: str) -> None:
        """Rename the product."""
        pass

    @abstractmethod
    def load_probe_positions(self, file_path: Path, *, file_type: str | None = None) -> None:
        """Load probe positions from file, uses format from settings when file_type is None."""
        pass

    @abstractmethod
    def generate_probe_positions(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        """Generate probe positions using the named generator and optional parameter overrides."""
        pass

    @abstractmethod
    def set_probe_positions_transform(self, transform: AffineTransform) -> None:
        """Apply an affine transform to the probe positions."""
        pass

    @abstractmethod
    def load_probe(self, file_path: Path, *, file_type: str | None = None) -> None:
        """Load a probe sequence from file, uses format from settings when file_type is None."""
        pass

    @abstractmethod
    def generate_probe(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        """Generate a probe using the named generator and optional parameter overrides."""
        pass

    @abstractmethod
    def load_object(self, file_path: Path, *, file_type: str | None = None) -> None:
        """Load an object array from file, uses format from settings when file_type is None."""
        pass

    @abstractmethod
    def generate_object(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        """Generate an object using the named generator and optional parameter overrides."""
        pass

    @abstractmethod
    def get_reconstruct_input(self) -> ReconstructInput:
        """Assemble and return the reconstruction input for this product."""
        pass

    @abstractmethod
    def reconstruct_local(
        self,
        *,
        algorithm: str | None = None,
        output_product_file: Path | None = None,
        block: bool = False,
    ) -> WorkflowProductAPI:
        """Run reconstruction locally; returns a handle to the output product.

        Blocks until completion when block is True, otherwise returns immediately.
        """
        pass

    @abstractmethod
    def reconstruct_remote(self, *, algorithm: str | None = None) -> None:
        """Submit reconstruction to a remote compute resource."""
        pass

    @abstractmethod
    def train_reconstructor_local(
        self,
        input_path: Path,
        output_path: Path,
        *,
        algorithm: str | None = None,
        block: bool = False,
    ) -> None:
        """Train the reconstructor model locally from input_path and write to output_path.

        Blocks until completion when block is True, otherwise returns immediately.
        """
        pass

    @abstractmethod
    def train_reconstructor_remote(self, *, algorithm: str | None = None) -> None:
        """Submit reconstructor training to a remote compute resource."""
        pass

    @abstractmethod
    def export_training_data(self, file_path: Path, *, algorithm: str | None = None) -> None:
        """Export training data for the current reconstructor to file."""
        pass

    @abstractmethod
    def save_product(self, file_path: Path, *, file_type: str | None = None) -> None:
        """Save the product to file, uses format from settings when file_type is None."""
        pass


class WorkflowAPI(ABC):
    """Top-level API for loading data, managing products, and running reconstructions."""

    @abstractmethod
    def load_diffraction_data(
        self,
        file_path: Path,
        *,
        file_type: str | None = None,
        crop_center: CropCenter | None = None,
        crop_extent: ImageExtent | None = None,
        detector_extent: ImageExtent | None = None,
        process_patterns: bool = True,
        block: bool = False,
    ) -> WorkflowDiffractionAPI:
        """Load and assemble a diffraction dataset from file.

        Blocks until complete when block is True, otherwise returns immediately.
        """
        pass

    @abstractmethod
    def load_assembled_diffraction_data(self, file_path: Path) -> WorkflowDiffractionAPI:
        """Load a pre-assembled diffraction dataset from file."""
        pass

    @abstractmethod
    def register_product(self, product: Product) -> WorkflowProductAPI:
        """Register an existing Product object and return a handle to it."""
        pass

    @abstractmethod
    def load_product(self, file_path: Path, *, file_type: str | None = None) -> WorkflowProductAPI:
        """Load a product from file and return a handle to it."""
        pass

    @abstractmethod
    def create_product(
        self,
        name: str,
        *,
        comments: str = '',
        detector_distance_m: float | None = None,
        probe_energy_eV: float | None = None,  # noqa: N803
        probe_photon_count: float | None = None,
        exposure_time_s: float | None = None,
        mass_attenuation_m2_kg: float | None = None,
        tomography_angle_deg: float | None = None,
    ) -> WorkflowProductAPI:
        """Create a new product with optional metadata overrides and return a handle to it."""
        pass

    @abstractmethod
    def get_product(self, product_index: int) -> WorkflowProductAPI:
        """Return a handle to an already-registered product by index."""
        pass

    @abstractmethod
    def available_reconstructors(self) -> Iterator[str]:
        """Yield the names of all registered reconstructors."""
        pass

    @abstractmethod
    def save_settings(
        self, file_path: Path, change_path_prefix: PathPrefixChange | None = None
    ) -> None:
        """Save the current settings to file, optionally remapping path prefixes."""
        pass


class FileBasedWorkflow(ABC):
    """Plugin interface for file-based workflows."""

    @property
    @abstractmethod
    def is_watch_recursive(self) -> bool:
        """True if the data directory must be watched recursively."""
        pass

    @abstractmethod
    def get_watch_file_pattern(self) -> str:
        """UNIX-style filename pattern. For rules see fnmatch from Python standard library."""
        pass

    @abstractmethod
    def execute(self, api: WorkflowAPI, file_path: Path) -> None:
        """Execute the workflow using the provided WorkflowAPI."""
        pass



from pathlib import Path
from typing import Final
import logging

import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionFileWriter,
    DiffractionMetadata,
    SimpleDiffractionDataset,
    SimpleDiffractionArray,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

logger = logging.getLogger(__name__)


class NPYDiffractionFileIO(DiffractionFileReader, DiffractionFileWriter):
    SIMPLE_NAME: Final[str] = 'NPY'
    DISPLAY_NAME: Final[str] = 'NumPy Binary Files (*.npy)'

    def read(self, file_path: Path) -> DiffractionDataset:
        data = numpy.load(file_path)

        if data.ndim == 2:
            data = data[numpy.newaxis, :, :]

        num_patterns, detector_height, detector_width = data.shape

        metadata = DiffractionMetadata(
            num_patterns_per_array=[num_patterns],
            pattern_dtype=data.dtype,
            detector_extent=ImageExtent(detector_width, detector_height),
            file_path=file_path,
        )

        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        contents_tree.create_child(
            [file_path.stem, type(data).__name__, f'{data.dtype}{data.shape}']
        )

        array = SimpleDiffractionArray(
            label=file_path.stem,
            indexes=numpy.arange(num_patterns),
            patterns=data,
        )

        return SimpleDiffractionDataset(metadata, contents_tree, [array])

    def write(self, file_path: Path, dataset: DiffractionDataset) -> None:
        patterns = numpy.concatenate([array.get_patterns() for array in dataset])
        numpy.save(file_path, patterns)


def register_plugins(registry: PluginRegistry) -> None:
    npy_diffraction_file_io = NPYDiffractionFileIO()

    registry.diffraction_file_readers.register_plugin(
        npy_diffraction_file_io,
        simple_name=NPYDiffractionFileIO.SIMPLE_NAME,
        display_name=NPYDiffractionFileIO.DISPLAY_NAME,
    )
    registry.diffraction_file_writers.register_plugin(
        npy_diffraction_file_io,
        simple_name=NPYDiffractionFileIO.SIMPLE_NAME,
        display_name=NPYDiffractionFileIO.DISPLAY_NAME,
    )



from pathlib import Path
import argparse


class DirectoryType:
    def __init__(self, *, must_exist: bool) -> None:
        self._must_exist = must_exist

    def __call__(self, string: str) -> Path:
        path = Path(string)

        if self._must_exist and not path.is_dir():
            raise argparse.ArgumentTypeError(f'"{string}" is not a directory!')

        return path


def verify_all_arguments_parsed(parser: argparse.ArgumentParser, argv: list[str]) -> None:
    if argv:
        parser.error('unrecognized arguments: %s' % ' '.join(argv))



from __future__ import annotations
from typing import Final
import logging
import time

from scipy.sparse.linalg import lsmr, LinearOperator
import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.fluorescence import (
    ElementMap,
    FluorescenceDataset,
    FluorescenceEnhancingAlgorithm,
)
from ptychodus.api.object import ObjectPosition
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.product import Product

from .settings import FluorescenceSettings

logger = logging.getLogger(__name__)

__all__ = [
    'VSPIFluorescenceEnhancingAlgorithm',
]


class ArrayPatchInterpolator:
    def __init__(self, array: RealArrayType, point: ObjectPosition, shape: tuple[int, ...]) -> None:
        # top left corner of object support
        xmin = point.coordinate_x_px - shape[-1] / 2
        ymin = point.coordinate_y_px - shape[-2] / 2

        # whole components (pixel indexes)
        xmin_wh = int(xmin)
        ymin_wh = int(ymin)

        # fractional (subpixel) components
        xmin_fr = xmin - xmin_wh
        ymin_fr = ymin - ymin_wh

        # bottom right corner of object patch support
        xmax_wh = xmin_wh + shape[-1] + 1
        ymax_wh = ymin_wh + shape[-2] + 1

        # reused quantities
        xmin_fr_c = 1.0 - xmin_fr
        ymin_fr_c = 1.0 - ymin_fr

        # barycentric interpolant weights
        self._weight00 = ymin_fr_c * xmin_fr_c
        self._weight01 = ymin_fr_c * xmin_fr
        self._weight10 = ymin_fr * xmin_fr_c
        self._weight11 = ymin_fr * xmin_fr

        # extract patch support region from full object
        self._support = array[ymin_wh:ymax_wh, xmin_wh:xmax_wh]

    def get_patch(self) -> RealArrayType:
        """interpolate array support to extract patch"""
        patch = self._weight00 * self._support[:-1, :-1]
        patch += self._weight01 * self._support[:-1, 1:]
        patch += self._weight10 * self._support[1:, :-1]
        patch += self._weight11 * self._support[1:, 1:]
        return patch

    def accumulate_patch(self, patch: RealArrayType) -> None:
        """add patch update to array support"""
        self._support[:-1, :-1] += self._weight00 * patch
        self._support[:-1, 1:] += self._weight01 * patch
        self._support[1:, :-1] += self._weight10 * patch
        self._support[1:, 1:] += self._weight11 * patch


class VSPILinearOperator(LinearOperator):
    def __init__(self, product: Product) -> None:
        """
        M: number of XRF positions
        N: number of ptychography object pixels
        P: number of XRF channels

        A[M,N] * X[N,P] = B[M,P]
        """
        object_geometry = product.object_.get_geometry()
        M = len(product.probe_positions)  # noqa: N806
        N = object_geometry.height_px * object_geometry.width_px  # noqa: N806
        super().__init__(float, (M, N))
        self._product = product

    def _matvec(self, x: RealArrayType) -> RealArrayType:  # noqa: N803
        object_geometry = self._product.object_.get_geometry()
        object_array = x.reshape((object_geometry.height_px, object_geometry.width_px))
        AX = numpy.zeros(len(self._product.probe_positions))  # noqa: N806

        for index, (scan_point, probe) in enumerate(
            zip(self._product.probe_positions, self._product.probes)
        ):
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            probe_intensity = probe.get_intensity()
            psf = probe_intensity / probe_intensity.sum()
            interpolator = ArrayPatchInterpolator(object_array, object_point, psf.shape)
            AX[index] = numpy.sum(psf * interpolator.get_patch())

        return AX

    def _rmatvec(self, x: RealArrayType) -> RealArrayType:  # noqa: N803
        object_geometry = self._product.object_.get_geometry()
        object_array = numpy.zeros((object_geometry.height_px, object_geometry.width_px))

        for index, (scan_point, probe) in enumerate(
            zip(self._product.probe_positions, self._product.probes)
        ):
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            probe_intensity = probe.get_intensity()
            psf = probe_intensity / probe_intensity.sum()
            interpolator = ArrayPatchInterpolator(object_array, object_point, psf.shape)
            interpolator.accumulate_patch(x[index] * psf)

        HX = object_array.flatten()  # noqa: N806

        return HX


class VSPIFluorescenceEnhancingAlgorithm(FluorescenceEnhancingAlgorithm, Observable, Observer):
    SIMPLE_NAME: Final[str] = 'VSPI'
    DISPLAY_NAME: Final[str] = 'Virtual Single Pixel Imaging'

    def __init__(self, settings: FluorescenceSettings) -> None:
        super().__init__()
        self._settings = settings

        settings.vspi_damping_factor.add_observer(self)
        settings.vspi_max_iterations.add_observer(self)

    def enhance(self, dataset: FluorescenceDataset, product: Product) -> FluorescenceDataset:
        object_geometry = product.object_.get_geometry()
        e_cps_shape = object_geometry.height_px, object_geometry.width_px
        element_maps: list[ElementMap] = list()
        A = VSPILinearOperator(product)  # noqa: N806

        for emap in dataset.element_maps:
            logger.info(f'Enhancing "{emap.name}"...')
            tic = time.perf_counter()
            m_cps = emap.counts_per_second
            result = lsmr(
                A,
                m_cps.flatten(),
                damp=self._settings.vspi_damping_factor.get_value(),
                maxiter=self._settings.vspi_max_iterations.get_value(),
                show=True,
            )
            logger.debug(result)
            e_cps = result[0].reshape(e_cps_shape)
            emap_enhanced = ElementMap(emap.name, e_cps)
            toc = time.perf_counter()
            logger.info(f'Enhanced "{emap.name}" in {toc - tic:.4f} seconds.')

            element_maps.append(emap_enhanced)

        return FluorescenceDataset(
            element_maps=element_maps,
            counts_per_second_path=dataset.counts_per_second_path,
            channel_names_path=dataset.channel_names_path,
        )

    def get_damping_factor(self) -> float:
        return self._settings.vspi_damping_factor.get_value()

    def set_damping_factor(self, factor: float) -> None:
        self._settings.vspi_damping_factor.set_value(factor)

    def get_max_iterations(self) -> int:
        return self._settings.vspi_max_iterations.get_value()

    def set_max_iterations(self, number: int) -> None:
        self._settings.vspi_max_iterations.set_value(number)

    def _update(self, observable: Observable) -> None:
        if observable is self._settings.vspi_damping_factor:
            self.notify_observers()
        elif observable is self._settings.vspi_max_iterations:
            self.notify_observers()



from typing import Any, Generic, TypeVar
import logging

import numpy.typing
import numpy

from ptychodus.api.common import RealArrayType

__all__ = [
    'BarycentricArrayInterpolator',
    'BarycentricArrayStitcher',
    'NearestNeighborArrayInterpolator',
]

InexactDType = TypeVar('InexactDType', bound=numpy.inexact[Any])

logger = logging.getLogger(__name__)


def calculate_support_frac(x: float, n: int) -> tuple[slice, float]:
    lower = x - n / 2
    whole = int(lower)
    return slice(whole, whole + n + 1), lower - whole


class NearestNeighborArrayInterpolator(Generic[InexactDType]):
    def __init__(self, array: numpy.typing.NDArray[InexactDType]) -> None:
        super().__init__()
        self._array = array

    def get_patch(
        self, center_x: float, center_y: float, width: int, height: int
    ) -> numpy.typing.NDArray[InexactDType]:
        y_lower = int(center_y - height / 2)
        y_support = slice(y_lower, y_lower + height)
        logger.debug(f'{y_support=}')

        x_lower = int(center_x - width / 2)
        x_support = slice(x_lower, x_lower + width)
        logger.debug(f'{x_support=}')

        return self._array[..., y_support, x_support]


class BarycentricArrayInterpolator(Generic[InexactDType]):
    def __init__(self, array: numpy.typing.NDArray[InexactDType]) -> None:
        super().__init__()
        self._array = array

    def get_patch(
        self, center_x: float, center_y: float, width: int, height: int
    ) -> numpy.typing.NDArray[InexactDType]:
        x_support, x_frac = calculate_support_frac(center_x, width)
        y_support, y_frac = calculate_support_frac(center_y, height)

        # reused quantities
        x_frac_c = 1.0 - x_frac
        y_frac_c = 1.0 - y_frac

        # barycentric interpolant weights
        weight00 = y_frac_c * x_frac_c
        weight01 = y_frac_c * x_frac
        weight10 = y_frac * x_frac_c
        weight11 = y_frac * x_frac

        support = self._array[..., y_support, x_support]
        patch = weight00 * support[..., :-1, :-1]
        patch = patch + weight01 * support[..., :-1, 1:]
        patch = patch + weight10 * support[..., 1:, :-1]
        patch = patch + weight11 * support[..., 1:, 1:]
        return patch  # type: ignore


class BarycentricArrayStitcher(Generic[InexactDType]):
    def __init__(
        self, upper: numpy.typing.NDArray[InexactDType], lower: RealArrayType | None = None
    ) -> None:
        super().__init__()
        self._upper = upper
        self._lower = lower

        if lower is not None and upper.shape != lower.shape:
            raise ValueError(f'Mismatched array shapes! ({upper.shape} != {lower.shape})')

    def add_patch(
        self,
        center_x: float,
        center_y: float,
        value: numpy.typing.NDArray[InexactDType],
        weight: RealArrayType | None = None,
    ) -> None:
        if numpy.iscomplexobj(self._upper) != numpy.iscomplexobj(value):
            raise ValueError(f'Mismatched value dtypes! ({self._upper.dtype} != {value.dtype})')

        if weight is not None:
            if self._lower is None:
                raise ValueError('Provided weights without a lower array!')

            if value.shape != weight.shape:
                raise ValueError(f'Mismatched patch shapes! ({value.shape=} != {weight.shape=})')

        x_support, x_frac = calculate_support_frac(center_x, value.shape[-1])
        y_support, y_frac = calculate_support_frac(center_y, value.shape[-2])

        # reused quantities
        x_frac_c = 1.0 - x_frac
        y_frac_c = 1.0 - y_frac

        # barycentric interpolant weights
        weight00 = y_frac_c * x_frac_c
        weight01 = y_frac_c * x_frac
        weight10 = y_frac * x_frac_c
        weight11 = y_frac * x_frac

        # add patch update to upper array support
        uvalue = value if weight is None else weight * value
        usupport = self._upper[..., y_support, x_support]
        usupport[..., :-1, :-1] = usupport[..., :-1, :-1] + weight00 * uvalue
        usupport[..., :-1, 1:] = usupport[..., :-1, 1:] + weight01 * uvalue
        usupport[..., 1:, :-1] = usupport[..., 1:, :-1] + weight10 * uvalue
        usupport[..., 1:, 1:] = usupport[..., 1:, 1:] + weight11 * uvalue

        if self._lower is not None and weight is not None:
            # add patch update to lower array support
            lsupport = self._lower[..., y_support, x_support]
            lsupport[..., :-1, :-1] += weight00 * weight
            lsupport[..., :-1, 1:] += weight01 * weight
            lsupport[..., 1:, :-1] += weight10 * weight
            lsupport[..., 1:, 1:] += weight11 * weight

    def stitch(self) -> numpy.typing.NDArray[InexactDType]:
        if self._lower is None:
            return self._upper

        return numpy.divide(
            self._upper, self._lower, out=numpy.zeros_like(self._upper), where=(self._lower > 0)
        )



from __future__ import annotations
from typing import Final

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QAction

from ..model import ModelCore
from ..view import ViewCore
from .agent import AgentChatController, AgentController
from .automation import AutomationController
from .data import FileDialogFactory
from .diffraction import DiffractionController
from .globus import GlobusController
from .image import ImageController
from .memory import MemoryController
from .object import ObjectController
from .probe import ProbeController
from .product import ProductController
from .ptychi import PtyChiViewControllerFactory
from .ptychonn import PtychoNNViewControllerFactory
from .ptychopinn import PtychoPINNViewControllerFactory
from .processing import ProcessingController
from .probe_positions import ProbePositionsController
from .settings import SettingsController


class ControllerCore:
    ONE_HOUR_S: Final[int] = 3600

    def __init__(
        self, model: ModelCore, view: ViewCore, *, is_developer_mode_enabled: bool = False
    ) -> None:
        self.model = model
        self.view = view
        self._status_bar = view.statusBar()

        if self._status_bar is None:
            raise ValueError('QStatusBar is None!')

        self._memory_controller = MemoryController(model.memory_presenter, view.memory_widget)
        self._file_dialog_factory = FileDialogFactory()
        self._ptychi_view_controller_factory = PtyChiViewControllerFactory(
            model.ptychi_reconstructor_library
        )
        self._ptychonn_view_controller_factory = PtychoNNViewControllerFactory(
            model.ptychonn_reconstructor_library
        )
        self._ptychopinn_view_controller_factory = PtychoPINNViewControllerFactory(
            model.ptychopinn_reconstructor_library, self._file_dialog_factory
        )
        self._settings_controller = SettingsController(
            model.settings_registry,
            view.settings_view,
            view.settings_table_view,
            self._file_dialog_factory,
        )
        self._patterns_image_controller = ImageController(
            model.pattern_visualization_engine,
            view.patterns_image_view,
            self._status_bar,
            self._file_dialog_factory,
        )
        self._patterns_controller = DiffractionController(
            model.diffraction_core.detector_settings,
            model.diffraction_core.diffraction_settings,
            model.diffraction_core.pattern_sizer,
            model.diffraction_core.bad_pixels_provider,
            model.diffraction_core.diffraction_api,
            model.diffraction_core.dataset,
            model.metadata_presenter,
            model.product_core.product_repository,
            model.analysis_core.diffraction_simulator,
            view.patterns_view,
            self._patterns_image_controller,
            self._file_dialog_factory,
        )
        self._product_controller = ProductController.create_instance(
            model.diffraction_core.diffraction_api,
            model.product_core.product_repository,
            model.product_core.product_api,
            view.product_view,
            self._file_dialog_factory,
        )
        self._probe_positions_controller = ProbePositionsController(
            model.product_core.probe_positions_repository,
            model.product_core.probe_positions_api,
            view.probe_positions_view,
            view.probe_positions_plot_view,
            self._file_dialog_factory,
            is_developer_mode_enabled=is_developer_mode_enabled,
        )
        self._probe_image_controller = ImageController(
            model.probe_visualization_engine,
            view.probe_image_view,
            self._status_bar,
            self._file_dialog_factory,
        )
        self._probe_controller = ProbeController(
            model.product_core.probe_repository,
            model.product_core.probe_api,
            self._probe_image_controller,
            model.analysis_core.probe_propagator,
            model.analysis_core.probe_propagator_visualization_engine,
            model.analysis_core.exposure_analyzer,
            model.analysis_core.exposure_visualization_engine,
            model.fluorescence_core.enhancer,
            model.fluorescence_core.visualization_engine,
            view.probe_view,
            self._file_dialog_factory,
        )
        self._object_image_controller = ImageController(
            model.object_visualization_engine,
            view.object_image_view,
            self._status_bar,
            self._file_dialog_factory,
        )
        self._object_controller = ObjectController(
            model.product_core.object_repository,
            model.product_core.object_api,
            self._object_image_controller,
            model.analysis_core.fourier_ring_correlator,
            model.analysis_core.fourier_analyzer,
            model.analysis_core.fourier_real_space_visualization_engine,
            model.analysis_core.fourier_reciprocal_space_visualization_engine,
            model.analysis_core.xmcd_analyzer,
            model.analysis_core.xmcd_visualization_engine,
            view.object_view,
            self._file_dialog_factory,
        )
        self._processing_controller = ProcessingController(
            model.processing_core.algorithm_parameter,
            model.processing_core.processing_api,
            model.product_core.product_repository,
            model.globus_core,
            view.processing_view,
            view.processing_status_view,
            self._file_dialog_factory,
            [
                self._ptychi_view_controller_factory,
                self._ptychopinn_view_controller_factory,
                self._ptychonn_view_controller_factory,
            ],
        )
        self._globus_controller = GlobusController(
            model.globus_core.settings,
            model.globus_core.authorizer,
            model.globus_core.status_repository,
            view.globus_view,
            view.globus_status_view,
            self._file_dialog_factory,
        )
        self._automation_controller = AutomationController(
            model.automation_core.settings,
            model.automation_core.repository,
            model.automation_core.presenter,
            view.automation_view,
            self._file_dialog_factory,
        )
        self._agent_controller = AgentController(
            model.agent_core.settings, model.agent_core.presenter, view.agent_view
        )
        self._agent_chat_controller = AgentChatController(
            model.agent_core.chat_history, model.agent_core.presenter, view.agent_chat_view
        )

        self._one_second_counter = 0
        self._run_tasks_timer = QTimer()
        self._run_tasks_timer.timeout.connect(self._run_tasks)
        self._run_tasks_timer.start(1000)

        view.globus_action.setVisible(model.globus_core.is_supported)

        self._swap_central_widgets(view.patterns_action)
        view.patterns_action.setChecked(True)
        view.navigation_action_group.triggered.connect(
            lambda action: self._swap_central_widgets(action)
        )

        view.agent_action.setVisible(is_developer_mode_enabled)
        view.probe_positions_view.button_box.analyze_button.setEnabled(is_developer_mode_enabled)

    def show_main_window(self, window_title: str) -> None:
        self.view.setWindowTitle(window_title)
        self.view.show()

    def _swap_central_widgets(self, action: QAction | None) -> None:
        if action is None:
            raise ValueError('QAction is None!')

        index = action.data()
        self.view.left_panel.setCurrentIndex(index)
        self.view.right_panel.setCurrentIndex(index)

    def _run_tasks(self) -> None:
        self.model.run_tasks()
        self._memory_controller.run_tasks(self._one_second_counter)
        self._globus_controller.run_tasks(self._one_second_counter)
        self._one_second_counter += 1

        # counter value is not intended to be precise; protect from overflow
        if self._one_second_counter > ControllerCore.ONE_HOUR_S:
            self._one_second_counter -= ControllerCore.ONE_HOUR_S



from collections.abc import Iterator, Sequence


class PtyChiEnumerators:
    def __init__(self) -> None:
        try:
            from ptychi.api import (
                BatchingModes,
                Directions,
                ForwardModels,
                ImageGradientMethods,
                ImageIntegrationMethods,
                LossFunctions,
                NoiseModels,
                OPRWeightSmoothingMethods,
                Optimizers,
                OrthogonalizationMethods,
                PatchInterpolationMethods,
                PositionCorrectionTypes,
            )
        except ModuleNotFoundError:
            self._batching_modes: Sequence[str] = list()
            self._directions: Sequence[str] = list()
            self._forward_models: Sequence[str] = list()
            self._image_gradient_methods: Sequence[str] = list()
            self._image_integration_methods: Sequence[str] = list()
            self._loss_functions: Sequence[str] = list()
            self._noise_models: Sequence[str] = list()
            self._opr_weight_smoothing_methods: Sequence[str] = list()
            self._optimizers: Sequence[str] = list()
            self._orthogonalization_methods: Sequence[str] = list()
            self._patch_interpolation_methods: Sequence[str] = list()
            self._position_correction_types: Sequence[str] = list()
        else:
            self._batching_modes = [member.name for member in BatchingModes]
            self._directions = [member.name for member in Directions]
            self._forward_models = [member.name for member in ForwardModels]
            self._image_gradient_methods = [member.name for member in ImageGradientMethods]
            self._image_integration_methods = [member.name for member in ImageIntegrationMethods]
            self._loss_functions = [member.name for member in LossFunctions]
            self._noise_models = [member.name for member in NoiseModels]
            self._opr_weight_smoothing_methods = [
                member.name for member in OPRWeightSmoothingMethods
            ]
            self._optimizers = [member.name for member in Optimizers]
            self._orthogonalization_methods = [member.name for member in OrthogonalizationMethods]
            self._patch_interpolation_methods = [
                member.name for member in PatchInterpolationMethods
            ]
            self._position_correction_types = [member.name for member in PositionCorrectionTypes]

    def batching_modes(self) -> Iterator[str]:
        return iter(self._batching_modes)

    def directions(self) -> Iterator[str]:
        return iter(self._directions)

    def forward_models(self) -> Iterator[str]:
        return iter(self._forward_models)

    def image_gradient_methods(self) -> Iterator[str]:
        return iter(self._image_gradient_methods)

    def image_integration_methods(self) -> Iterator[str]:
        return iter(self._image_integration_methods)

    def loss_functions(self) -> Iterator[str]:
        return iter(self._loss_functions)

    def noise_models(self) -> Iterator[str]:
        return iter(self._noise_models)

    def opr_weight_smoothing_methods(self) -> Iterator[str]:
        return iter(self._opr_weight_smoothing_methods)

    def optimizers(self) -> Iterator[str]:
        return iter(self._optimizers)

    def orthogonalization_methods(self) -> Iterator[str]:
        return iter(self._orthogonalization_methods)

    def patch_interpolation_methods(self) -> Iterator[str]:
        return iter(self._patch_interpolation_methods)

    def position_correction_types(self) -> Iterator[str]:
        return iter(self._position_correction_types)



from __future__ import annotations
from abc import abstractmethod
from collections.abc import Sequence
from pathlib import Path
import logging

import numpy

from ptychodus.api.common import ComplexArrayType
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import Object, ObjectCenter, ObjectFileReader, ObjectGeometryProvider
from ptychodus.api.parametric import ParameterGroup

from ...phase_unwrapper import PhaseUnwrapper
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class ObjectBuilder(ParameterGroup):
    def __init__(self, settings: ObjectSettings, name: str) -> None:
        super().__init__()
        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

        self.extra_padding_x = settings.extra_padding_x.copy()
        self._add_parameter('extra_padding_x', self.extra_padding_x)
        self.extra_padding_y = settings.extra_padding_y.copy()
        self._add_parameter('extra_padding_y', self.extra_padding_y)

    def get_name(self) -> str:
        return self._name.get_value()

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ObjectBuilder:
        pass

    @abstractmethod
    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        pass

    def _create_object(
        self,
        array: ComplexArrayType,
        pixel_geometry: PixelGeometry,
        center: ObjectCenter,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        """Create an object from an existing object with a potentially
        different number of slices.

        If the new object is supposed to be a multislice object with a
        different number of slices than the existing object, the object is
        created as
        `abs(o) ** (1 / nSlices) * exp(i * unwrapPhase(o) / nSlices)`.
        Otherwise, the object is copied as is.
        """
        num_slices = 1 + len(layer_spacing_m)

        if array.ndim < 2:
            raise ValueError('Array must have at least 2 dimensions')
        elif array.ndim == 2:
            array = numpy.expand_dims(array, axis=0)
        elif array.ndim > 3:
            raise ValueError('Array must have at most 3 dimensions')

        if num_slices < array.shape[0]:
            array = array[:num_slices]
        elif num_slices > array.shape[0]:
            amplitude = numpy.absolute(array[:1]) ** (1.0 / num_slices)
            amplitude = amplitude.repeat(num_slices, axis=0)
            phase = PhaseUnwrapper().unwrap(array[0])[numpy.newaxis, ...] / num_slices
            phase = phase.repeat(num_slices, axis=0)
            array = numpy.clip(amplitude, 0.0, 1.0) * numpy.exp(1j * phase)

        pad_width = [
            (0, 0),
            (self.extra_padding_y.get_value(), self.extra_padding_y.get_value()),
            (self.extra_padding_x.get_value(), self.extra_padding_x.get_value()),
        ]

        return Object(
            array=numpy.pad(array, pad_width),
            layer_spacing_m=layer_spacing_m,
            pixel_geometry=pixel_geometry,
            center=center,
        )


class FromMemoryObjectBuilder(ObjectBuilder):
    def __init__(self, settings: ObjectSettings, object_: Object) -> None:
        super().__init__(settings, 'from_memory')
        self._settings = settings
        self._object = object_.copy()

    def copy(self) -> FromMemoryObjectBuilder:
        return FromMemoryObjectBuilder(self._settings, self._object)

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_geometry = geometry_provider.get_object_geometry()

        try:
            pixel_geometry = self._object.get_pixel_geometry()
        except ValueError:
            pixel_geometry = object_geometry.get_pixel_geometry()

        try:
            center = self._object.get_center()
        except ValueError:
            center = object_geometry.get_center()

        return Object(
            self._object.get_array(),
            pixel_geometry,
            center,
            self._object.layer_spacing_m,
        )


class FromFileObjectBuilder(ObjectBuilder):
    def __init__(
        self,
        settings: ObjectSettings,
        file_path: Path,
        file_type: str,
        file_reader: ObjectFileReader,
    ) -> None:
        super().__init__(settings, 'from_file')
        self._settings = settings
        self.file_path = settings.file_path.copy()
        self.file_path.set_value(file_path)
        self._add_parameter('file_path', self.file_path)
        self.file_type = settings.file_type.copy()
        self.file_type.set_value(file_type)
        self._add_parameter('file_type', self.file_type)
        self._file_reader = file_reader

    def copy(self) -> FromFileObjectBuilder:
        return FromFileObjectBuilder(
            self._settings,
            self.file_path.get_value(),
            self.file_type.get_value(),
            self._file_reader,
        )

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            object_from_file = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        object_geometry = geometry_provider.get_object_geometry()

        try:
            pixel_geometry = object_from_file.get_pixel_geometry()
        except ValueError:
            pixel_geometry = object_geometry.get_pixel_geometry()

        try:
            center = object_from_file.get_center()
        except ValueError:
            center = object_geometry.get_center()

        return Object(
            object_from_file.get_array(),
            pixel_geometry,
            center,
            object_from_file.layer_spacing_m,
        )



from dataclasses import dataclass
import logging

from scipy.fft import fft2, fftshift, ifftshift

from ptychodus.api.common import ComplexArrayType
from ptychodus.api.geometry import Box2D, PixelGeometry
from ptychodus.api.object import Object
from ptychodus.api.observer import Observable

from ..product import ProductRepository
from .interpolators import NearestNeighborArrayInterpolator


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class FourierAnalysisResult:
    transformed_roi: ComplexArrayType
    pixel_geometry: PixelGeometry


class FourierAnalyzer(Observable):
    def __init__(self, repository: ProductRepository) -> None:
        super().__init__()
        self._repository = repository
        self._product_index = -1
        self._result: FourierAnalysisResult | None = None

    def set_product(self, product_index: int) -> None:
        if self._product_index != product_index:
            self._product_index = product_index
            object_ = self.get_object()
            self._analyze(object_.get_layer(0))

    def get_product_name(self) -> str:
        product = self._repository[self._product_index]
        return product.get_name()

    def get_object(self) -> Object:
        product = self._repository[self._product_index]
        return product.get_object_item().get_object()

    def _analyze(self, array: ComplexArrayType) -> None:
        norm = 'forward'  # TODO let user choose norm
        object_ = self.get_object()
        self._result = FourierAnalysisResult(
            transformed_roi=fftshift(fft2(ifftshift(array), norm=norm)),
            pixel_geometry=object_.get_pixel_geometry(),
        )
        self.notify_observers()

    def analyze_roi(self, bounding_box: Box2D) -> None:
        logger.debug(f'bounding_box: {bounding_box}')
        object_ = self.get_object()
        interpolator = NearestNeighborArrayInterpolator(object_.get_layer(0))

        width = int(bounding_box.width + 0.5)
        height = int(bounding_box.height + 0.5)
        roi = interpolator.get_patch(bounding_box.x_center, bounding_box.y_center, width, height)
        logger.debug(f'roi: {roi.dtype}{roi.shape}')
        self._analyze(roi)

    def get_result(self) -> FourierAnalysisResult:
        if self._result is None:
            raise ValueError('Fourier analysis has not been performed yet.')

        return self._result



import logging

from ptychodus.api.plugins import PluginChooser
from ptychodus.api.product import Product, ProductFileReader

from ..diffraction import AssembledDiffractionDataset, PatternSizer
from .geometry import ProductGeometry
from .item import ProductRepositoryItem
from .metadata import MetadataRepositoryItem
from .object import ObjectRepositoryItemFactory
from .probe import ProbeRepositoryItemFactory
from .repository import ProductRepository
from .probe_positions import ProbePositionsRepositoryItemFactory
from .settings import ProductSettings
from .validator import ProductValidator

logger = logging.getLogger(__name__)


class ProductRepositoryItemFactory:
    def __init__(
        self,
        settings: ProductSettings,
        pattern_sizer: PatternSizer,
        dataset: AssembledDiffractionDataset,
        scan_item_factory: ProbePositionsRepositoryItemFactory,
        probe_item_factory: ProbeRepositoryItemFactory,
        object_item_factory: ObjectRepositoryItemFactory,
        repository: ProductRepository,
        file_reader_chooser: PluginChooser[ProductFileReader],
    ) -> None:
        super().__init__()
        self._settings = settings
        self._pattern_sizer = pattern_sizer
        self._dataset = dataset
        self._scan_item_factory = scan_item_factory
        self._probe_item_factory = probe_item_factory
        self._object_item_factory = object_item_factory
        self._repository = repository
        self._file_reader_chooser = file_reader_chooser

    def create_from_values(
        self,
        *,
        name: str = '',
        comments: str = '',
        detector_distance_m: float | None = None,
        probe_energy_eV: float | None = None,  # noqa: N803
        probe_photon_count: float | None = None,
        exposure_time_s: float | None = None,
        mass_attenuation_m2_kg: float | None = None,
        tomography_angle_deg: float | None = None,
    ) -> ProductRepositoryItem:
        metadata_item = MetadataRepositoryItem(
            self._settings,
            self._repository,
            name=name,
            comments=comments,
            detector_distance_m=detector_distance_m,
            probe_energy_eV=probe_energy_eV,
            probe_photon_count=probe_photon_count,
            exposure_time_s=exposure_time_s,
            mass_attenuation_m2_kg=mass_attenuation_m2_kg,
            tomography_angle_deg=tomography_angle_deg,
        )

        if metadata_item.probe_photon_count.get_value() <= 0:
            assembled_data = self._dataset.get_assembled_data()
            max_pattern_counts = assembled_data.get_pattern_counts().max()
            metadata_item.probe_photon_count.set_value(max_pattern_counts)

        scan_item = self._scan_item_factory.create()
        geometry = ProductGeometry(self._pattern_sizer, metadata_item, scan_item)
        probe_item = self._probe_item_factory.create(geometry)
        object_item = self._object_item_factory.create(geometry)
        validator = ProductValidator(self._dataset, scan_item, geometry, probe_item, object_item)

        return ProductRepositoryItem(
            parent=self._repository,
            metadata_item=metadata_item,
            probe_positions_item=scan_item,
            geometry=geometry,
            probe_item=probe_item,
            object_item=object_item,
            validator=validator,
            losses=list(),
        )

    def create_from_product(self, product: Product) -> ProductRepositoryItem:
        metadata_item = MetadataRepositoryItem(
            self._settings,
            self._repository,
            name=product.metadata.name,
            comments=product.metadata.comments,
            detector_distance_m=product.metadata.detector_distance_m,
            probe_energy_eV=product.metadata.probe_energy_eV,
            probe_photon_count=product.metadata.probe_photon_count,
            exposure_time_s=product.metadata.exposure_time_s,
            mass_attenuation_m2_kg=product.metadata.mass_attenuation_m2_kg,
            tomography_angle_deg=product.metadata.tomography_angle_deg,
        )

        scan_item = self._scan_item_factory.create(product.probe_positions)
        geometry = ProductGeometry(self._pattern_sizer, metadata_item, scan_item)
        probe_item = self._probe_item_factory.create(geometry, product.probes)
        object_item = self._object_item_factory.create(geometry, product.object_)
        validator = ProductValidator(self._dataset, scan_item, geometry, probe_item, object_item)

        return ProductRepositoryItem(
            parent=self._repository,
            metadata_item=metadata_item,
            probe_positions_item=scan_item,
            geometry=geometry,
            probe_item=probe_item,
            object_item=object_item,
            validator=validator,
            losses=product.losses,
        )

    def create_from_settings(self) -> ProductRepositoryItem:
        file_path = self._settings.file_path.get_value()

        if file_path.is_file():
            file_type = self._file_reader_chooser.get_current_plugin().simple_name
            logger.debug(f'Reading "{file_path}" as "{file_type}"')
            file_reader = self._file_reader_chooser.get_current_plugin().strategy

            try:
                product = file_reader.read(file_path)
            except Exception as exc:
                raise RuntimeError(f'Failed to read "{file_path}"') from exc
            else:
                return self.create_from_product(product)

        metadata_item = MetadataRepositoryItem(self._settings, self._repository)
        scan_item = self._scan_item_factory.create_from_settings()
        geometry = ProductGeometry(self._pattern_sizer, metadata_item, scan_item)
        probe_item = self._probe_item_factory.create_from_settings(geometry)
        object_item = self._object_item_factory.create_from_settings(geometry)

        item = ProductRepositoryItem(
            parent=self._repository,
            metadata_item=metadata_item,
            probe_positions_item=scan_item,
            geometry=geometry,
            probe_item=probe_item,
            object_item=object_item,
            validator=ProductValidator(self._dataset, scan_item, geometry, probe_item, object_item),
            losses=list(),
        )
        logger.debug(f'Created product from settings: {item.get_name()}')
        return item



from PyQt5.QtWidgets import (
    QFormLayout,
    QGroupBox,
    QPushButton,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import BooleanParameter, IntegerParameter

from ...controller.parametric import SpinBoxParameterViewController
from ...model.globus import GlobusStatusRepository
from ..parametric import CheckBoxParameterViewController, ParameterViewController


class GlobusStatusViewController(ParameterViewController, Observer):
    def __init__(
        self,
        status_auto_refresh: BooleanParameter,
        status_refresh_interval_s: IntegerParameter,
        status_repository: GlobusStatusRepository,
    ) -> None:
        super().__init__()
        self._status_auto_refresh = status_auto_refresh
        self._status_refresh_interval_s = status_refresh_interval_s
        self._status_repository = status_repository

        self._auto_refresh_view_controller = CheckBoxParameterViewController(
            status_auto_refresh, 'Auto Refresh [sec]:'
        )
        self._status_refresh_interval_view_controller = SpinBoxParameterViewController(
            status_refresh_interval_s
        )
        self._refresh_button = QPushButton('Refresh')
        self._refresh_button.clicked.connect(status_repository.refresh_status)

        layout = QFormLayout()
        layout.addRow(
            self._auto_refresh_view_controller.get_widget(),
            self._status_refresh_interval_view_controller.get_widget(),
        )
        layout.addRow(self._refresh_button)

        self._widget = QGroupBox('Status')
        self._widget.setLayout(layout)

        status_auto_refresh.add_observer(self)
        self._update_enabled_widgets()

    def get_widget(self) -> QWidget:
        return self._widget

    def _update_enabled_widgets(self) -> None:
        status_refresh_interval_widget = self._status_refresh_interval_view_controller.get_widget()
        enable = not self._status_auto_refresh.get_value()
        status_refresh_interval_widget.setEnabled(enable)
        self._refresh_button.setEnabled(enable)

    def _update(self, observable: Observable) -> None:
        if observable is self._status_auto_refresh:
            self._update_enabled_widgets()
